#------------------------------------------------------------------
# $Copyright: Copyright (c) 2011 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/SAPWebAS71/SAPWebAS71.pm,v 1.28 2011/06/14 07:44:22 lpatil Exp $
#
# $Author: lpatil $
# $Date: 2011/06/14 07:44:22 $
# $Revision: 1.28 $
#
# Provides:                            SAPWebAS71
# Author:                              vchappav
# crAgent Build:                       1.88 2008/10/13 08:28:02
# crAgent Run Date:                    13 Oct, 2008
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package SAPWebAS71;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use POSIX qw ( uname );
use POSIX qw ( strftime );
use File::Path qw ( mkpath );
use File::stat;

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vcs(|one))$/;

if ( $sPrefix =~ /vcsone/ ) {
   $sPrefix = "ha";
}

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHagrp  = $HAGRP  ||= "$CLUS_HOME/bin/hagrp";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdHasys  = $HASYS  ||= "$CLUS_HOME/bin/hasys";
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: SAPWebAS71.pm,v 1.28 2011/06/14 07:44:22 lpatil Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#-----------------
# Define globals..
#-----------------
my ( $iSapStartSrv, $iCCMSX, $iCCMSR, $bSapStart, @liRunningProcs ) = (); # Global..

#--------------------------------
# Define SAPWebAS71 subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String EnvFile,
                              IN String SAPAdmin,
                              IN String InstProfile,
                              IN String InstType,
                              IN String ProcMon,
                              IN String EnqSrvResName,
                              IN Integer SecondLevelMonitor,
                              IN String MonitorProgram );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry point name
 $localEnvFile                  Absolute path to environment file to source
 $localSAPAdmin                 SAP System Administrator
 $localInstProfile              Instance profile file for the SAP Instance
 $localInstType                 SAP Instance Type
 $localProcMon                  List of processes to be monitored
 $localEnqSrvResName            VCS resource name for Enqueue Server
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localMonitorProgram           Full path to external monitor program
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localEnvFile,
       $localSAPAdmin,
       $localInstProfile,
       $localInstType,
       $localProcMon,
       $localEnqSrvResName,
       $localSecondLevelMonitor,
       $localMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  SAPAdmin                 [%s]
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  EnqSrvResName            [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]",

  $sMyName,   $sEntryPointName,   $localEnvFile,    $localSAPAdmin,
  $localInstProfile,         $localInstType,         $localProcMon,
  $localEnqSrvResName,                    $localSecondLevelMonitor,
  $localMonitorProgram );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName  EnvFile  SAPAdmin  InstProfile  InstType  ProcMon
       EnqSrvResName SecondLevelMonitor MonitorProgram );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( EnvFile SAPAdmin InstProfile InstType ProcMon );
  my @lsOfflineArgs = qw ( SAPAdmin InstProfile InstType );
  my @lsMonitorArgs = qw ( SAPAdmin InstProfile InstType ProcMon );
  my @lsCleanArgs   = qw ( SAPAdmin InstProfile InstType );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  my ($sSAPSID, $sInstName, $sSAPHost) = ();
  my (@lsWords) = ();

  (@lsWords) = split( /\//, $localInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);

  #-------------------------------------------------------
  # Validation for online offline monitor entry point(s)..
  #-------------------------------------------------------
  if ( $sEntryPointName =~ /^(online|offline|monitor)$/o ) {
    #------------------------------
    # Validate parameter SAPAdmin..
    #------------------------------
    if ( ! defined getpwnam $localSAPAdmin ) {
      &logError ( 20003, "<%s> SAPAdmin [%s] is not a valid UNIX login",
      $sMyName, $localSAPAdmin );
      $iReturnCode = 0; # Failure..
    }
  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #-----------------------------
    # Validate parameter EnvFile..
    #-----------------------------
    if ( ! -e $localEnvFile ) {
      &logError ( 20004, "<%s> EnvFile [%s] file does not exist",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }

    #---------------------------------------
    # Validate SAPHost..
    #---------------------------------------
    if ( ! inet_aton ( $sSAPHost ) ) {
      &logError ( 20005,
      "<%s> InstProfile [%s] not terminated with valid hostname [%s]",
      $sMyName, $localInstProfile, $sSAPHost );
      $iReturnCode = 0; # Failure..
    }

    #---------------------------------
    # Validate parameter InstProfile..
    #---------------------------------
    if ( ! -e $localInstProfile ) {
      &logError ( 20006, "<%s> InstProfile [%s] file does not exist",
      $sMyName, $localInstProfile );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20007, "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }

    #----------------------------------
    # Check if EnqSrvResName is valid..
    #----------------------------------
    if ( $localEnqSrvResName =~ m/\S+/ && $localInstType =~ /^enqrep$/io  &&
         ! SAPWebAS71::ChkResName ( $localEnqSrvResName,
                                    $sSAPSID,
                                    $localInstType ) ) {
      &logError ( 20008,
      "<%s> EnqSrvResName [%s] invalid resource of Type [SAPWebAS71]",
      $sMyName, $localEnqSrvResName );
      $iReturnCode = 0; # Failure..
    }
  }

  #----------------------------------------------
  # Validate attributes for online and monitor ..
  #----------------------------------------------
  if ( $sEntryPointName =~ /^(online|monitor)$/o ) {
    #---------------------------------------------
    # Validate ProcMon after removing duplicates..
    #--------------------------------------------
    my $rslocalProcMon = \$_[6]; # Get reference to what's passed in..
    my %hsProcMon = map { $_ => 1 } $localProcMon =~ /([^\s]+)/g;
    my @lsProcMon = keys %hsProcMon;
    my ( @lsBadProcs, %hsAllowed ) = ();

    if ( $localInstType =~ /^(appserv)$/io ) {
      if ( ! grep { /^(dw|jc|jstart)$/o } @lsProcMon ) { # Must have..
        &logError ( 20009, "<%s> Incorrect value for ProcMon [%s].
        Has to include at least one of [dw jc jstart] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }

      %hsAllowed = map { $_ => 1 } qw ( dw co se jc jstart ig gw cm cs gwrd icman ); # Restricted set..
      if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
        &logError ( 20010, "<%s> Incorrect value(s) [%s] in ProcMon found.
        Must be one of [%s] for InstType [%s]", $sMyName,
        join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ),
        $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }
    elsif ( $localInstType =~ /^(enqueue)$/io ) {
      if ( ! grep { /^en$/o } @lsProcMon ) {
        &logError ( 20011, "<%s> Incorrect value for ProcMon [%s].
        Has to include at least [en] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }

      %hsAllowed = map { $_ => 1 } qw ( en ms gw ); # Restricted set..
      if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
        &logError ( 20012, "<%s> Incorrect value(s) [%s] in ProcMon found.
        Must be one of [%s] for InstType [%s]", $sMyName,
        join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ),
        $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }
    elsif ( $localInstType =~ /^(enqrep)$/io  ) {
      if ( ! grep { /^(er)$/o } @lsProcMon ) {
        &logError ( 20013, "<%s> Incorrect value for ProcMon [%s].
        Must be [er] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }
    elsif( $localInstType =~ /^(SMDAgent)$/io ) {
      if ( ! grep { /^(jc|jstart)$/o } @lsProcMon ) { # Must have..
        &logError ( 20327, "<%s> Incorrect value for ProcMon [%s].
        Has to include at least one of [dw jc jstart] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }


    #--------------------
    # Re-create procmon..
    #--------------------
    $$rslocalProcMon = join ( ' ', @lsProcMon );
  }

  #-------------------------------------------
  # Validate attributes for all entry points..
  #-------------------------------------------

  #-----------------------------------------
  # SAPSID.. must be three characters long..
  #-----------------------------------------
  if ( $sSAPSID !~ /^\S{3}$/o ) {
    &logError ( 20014,
    "<%s> SAPSID [%s] has to be exactly 3 non-white alpha-numeric characters long",
    $sMyName, $sSAPSID );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------------------
  # InstName. Last two chars 00 - 99..
  #-----------------------------------
  unless ( $sInstName =~ /(\d{2})$/o ) {
    &logError ( 20015,
    "<%s> Last two characters of InstName [%s] not between 00-99",
    $sMyName, $sInstName );
    $iReturnCode = 0; # Failure..
  }

  #------------------
  # Valid InstTypes..
  #------------------
  my @lsIntTypes = qw ( appserv enqueue enqrep smdagent  );

  unless ( grep { /^$localInstType$/io } @lsIntTypes ) {
    &logError ( 20016,
    "<%s> InstType [%s] invalid. Has to be one of [%s]",
    $sMyName, $localInstType, join ( ' ', @lsIntTypes ) );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode ||
  &logWarn ( 20017,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String EnvFile,
                         IN String InstProfile,
                         IN String InstType,
                         IN String ProcMon,
                         IN String SAPAdmin,
                         IN String SAPSID,
                         IN String SAPHome,
                         IN String InstName,
                         IN Integer SecondLevelMonitor );

 Description : Start logic needed for online entry pt

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $sSAPSID                   SAP System ID
 $sSAPHome                  SAP Instance Home
 $sInstName                 SAP Instance Name
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin,
       $sSAPSID,
       $sSAPHome,
       $sInstName,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20018, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  SAPHome                  [%s]
  InstName                 [%s]
  SecondLevelMonitor       [%s]",

  $sMyName, $sEnvFile, $sInstProfile, $sInstType, $sProcMon,
  $sSAPAdmin, $sSAPSID, $sSAPHome, $sInstName, $iSecondLevelMonitor );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #-----------------------------------------
  # Whack any preexisting SAPWebAS71 processes..
  #-----------------------------------------
  &logTrace ( 20019,
  "<%s> Killing SAPWebAS71 processes for this instance", $sMyName );

  if ( SAPWebAS71::KillSAPWebAS71Processes ( $sInstProfile,
                                             $sInstType,
                                             $sProcMon,
                                             $sSAPAdmin ) ) {
    &logTrace ( 20020,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20021,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPWebAS71::GetEngineTimeout ( 1 );

  my $iInstID = ();

  #-----------------------------------
  # Check if it's safe to clean IPCs..
  #-----------------------------------
  if ( $sInstType =~ /^(appserv|enqrep)$/io ) {
    #--------------------------------------------------------
    # Clean IPC resources for the instance ..
    #--------------------------------------------------------
    ($iInstID ) = $sInstName =~ /(\d{2})$/o;

    $sCommand = "$sSAPHome/cleanipc";
    $sCommandArgs = "$iInstID remove";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();

    &logTrace ( 20022,
    "<%s> Cleaning IPC resources for SAPWebAS71 instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );

    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20023,
      "<%s> Insufficient time to run cleanipc. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
      SAPWebAS71::ExecLocalCmd ( $sSAPAdmin, 
                                 $sEnvFile, 
                                 $sCommand, 
                                 $sCommandArgs, 
                                 \$iExitCode, 
                                 $iTimeout, 
                                 \@lsCmdout ) ) {
      &logTrace ( 20024,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }
  else {
    &logTrace ( 20025,
    "<%s> Skipping cleanipc for InstType [%s].
    Possibility of destroying active lock table", $sMyName, $sInstType );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  my ($sKillFile, $sShutdownFile, $sCollLockFile, $sSendLockFile, @lsFiles ) = ();

  #------------------
  # Remove kill.sap..
  #------------------
  $sKillFile = "/usr/sap/$sSAPSID/$sInstName/work/kill.sap";

  if ( -e $sKillFile ) {
    &logTrace ( 20026,
    "<%s> Located file [%s] for removal", $sMyName, $sKillFile );
    push ( @lsFiles, $sKillFile );
  }

  #----------------------
  # Remove shutdown.sap..
  #----------------------
  $sShutdownFile = "/usr/sap/$sSAPSID/$sInstName/work/shutdown.sap";

  if ( -e $sShutdownFile ) {
    &logTrace ( 20027,
    "<%s> Located file [%s] for removal", $sMyName, $sShutdownFile );
    push ( @lsFiles, $sShutdownFile );
  }

  #-----------------------------
  # Remove collector lock file..
  #-----------------------------
  $sCollLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgcpid";

  if ( -e $sCollLockFile ) {
    &logTrace ( 20028,
    "<%s> Located file [%s] for removal", $sMyName, $sCollLockFile );
    push ( @lsFiles, $sCollLockFile );
  }

  #--------------------------
  # Remove sender lock file..
  #--------------------------
  $sSendLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgspid";

  if ( -e $sSendLockFile ) {
    &logTrace ( 20029,
    "<%s> Located file [%s] for removal", $sMyName, $sSendLockFile );
    push ( @lsFiles, $sSendLockFile );
  }

  #-------------------
  # Remove the files..
  #-------------------
  if ( scalar @lsFiles && ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
    &logError ( 20030, "<%s> Failed to remove stale file(s) [%s]",
    $sMyName, join ( ' ', @lsFiles ) );
  }

  #----------------------------------
  # Initiate SAPWebAS71 startup log..
  #----------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sMessage = "Starting up [$sResourceName]";
  if ( ! ( $sInstType =~ /smdagent/io ) ) {
    if ( SAPWebAS71::UpdateSAPWebAS71Log ( $sInstName,
                                           $sInstProfile,
                                           $sInstType,
                                           $sSAPAdmin,
                                           $sSAPSID,
                                           $sMessage ) ) {
      &logTrace ( 20031,
      "<%s> Updated log successfully", $sMyName );
    }
    else {
      &logError ( 20032,
      "<%s> Could not log successfully", $sMyName );
    }
  }

  #------------------------------------------------
  # Define sapstartsrv command and args variables..
  #------------------------------------------------
  $sCommand = "$sSAPHome/sapstartsrv";
  $sCommandArgs = "pf=$sInstProfile -D -u $sSAPAdmin";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20033,
  "<%s> Starting SAPWebAS71 instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20034,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    SAPWebAS71::ExecLocalCmd ( $sSAPAdmin, 
                               $sEnvFile, 
                               $sCommand, 
                               $sCommandArgs, 
                               \$iExitCode, 
                               $iTimeout, 
                               \@lsCmdout ) ) {
    &logTrace ( 20035,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #------------------------------------------
  # Start SAPWebAS71 Instance ..
  #------------------------------------------
  my %hsExitCode = ( 0 => 'Success',
                     1 => 'OS configuration error',
                     2 => 'Called with wrong Parameters',
                     3 => 'User not authorized to execute script',
                     4 => 'Environment not set',
                     5 => 'Database startup failed',
                     6 => 'Database on remote DB-Server not running',
                     7 => 'Startup of Instance failed',
                     8 => 'Start profile not found',
                     9 => 'User not authorized to start database' );

  my $sLogFile = (); #"/usr/sap/$sSAPSID/$sInstName/log/" .  "${sSAPSID}_${sInstName}.log"; 
  if ( $sInstType =~ /smdagent/io ) {
    $sLogFile = "/dev/null";
  }
  else {
    $sLogFile = "/usr/sap/$sSAPSID/$sInstName/log/" . "${sSAPSID}_${sInstName}.log";
  }
  #-------------------------------------------
  # Get re-direct string based on user shell..
  #-------------------------------------------
  my $sRdStr = ( getpwnam $sSAPAdmin ) [-1] =~ /csh$/o ? ">>& $sLogFile":
                                                         "1>> $sLogFile 2>&1";
  
  $sCommand = "$sSAPHome/sapstart";
  $sCommandArgs = "pf=$sInstProfile $sRdStr";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20036,
  "<%s> Starting SAPWebAS71 instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20037,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    SAPWebAS71::ExecLocalCmd ( $sSAPAdmin, 
                               $sEnvFile, 
                               $sCommand, 
                               $sCommandArgs, 
                               \$iExitCode, 
                               $iTimeout, 
                               \@lsCmdout ) ) {
    &logTrace ( 20038,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20039,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         SAPWebAS71::GetProcState ( \@liPids,
                                    $sInstProfile,
                                    $sInstType,
                                    $sProcMon,
                                    $sSAPAdmin,
                                    1 ) ) {          # Running FFLM..
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20040,
      "<%s> Process check confirmed start-up of this instance of SAPWebAS71 server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20041,
      "<%s> Process check could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }

  if ( $sInstType !~ /^(enqueue|enqrep)$/io && $iSecondLevelMonitor ) {      # If enabled..
    while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
      @liPids = (); # Initialise..

      sleep $iSleepTime; # Sec..

      &logTrace ( 20042,
      "<%s> Confirming start-up with [%s] sec in hand by health check",
      $sMyName, $iTimeout - $iSleepTime );

      #---------
      # Do SLM..
      #---------
      if ( VCSDef_MonitorExit::ONLINE ==                   # Online after..
           SAPWebAS71::SecondLevelMonitor ( $sEnvFile,     # Running SLM..
                                            $sProcMon,
                                            $sInstProfile,
                                            $sInstType,
                                            $sSAPAdmin,
                                            $iTillWhen ) ) { 
        #---------------------
        # Start-up confirmed..
        #---------------------
        &logInfo ( 20043,
        "<%s> Second Level Monitor confirmed start-up of this instance of SAPWebAS71 server",
        $sMyName );

        $iReturnCode = 1; # Success..
        last;
      }
      else {
        &logTrace ( 20044,
        "<%s> Second Level Monitor could not confirm start-up. Sleeping for [%s] sec",
        $sMyName, $iSleepTime );

        $iReturnCode = 0; # Failure..
      }
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String EnvFile,
                        IN String InstProfile,
                        IN String InstType,
                        IN String ProcMon,
                        IN String SAPAdmin,
                        IN String SAPSID,
                        IN String SAPHome,
                        IN String InstName );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $sSAPSID                   SAP System ID
 $sSAPHome                  SAP Instance Home
 $sInstName                 SAP Instance Name
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin, 
       $sSAPSID, 
       $sSAPHome, 
       $sInstName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20045, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  SAPHome                  [%s]
  InstName                 [%s]",

  $sMyName, $sEnvFile, $sInstProfile, $sInstType, $sProcMon, 
  $sSAPAdmin, $sSAPSID, $sSAPHome, $sInstName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPWebAS71::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  
  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  my $sSAPWork = "/usr/sap/$sSAPSID/$sInstName/work" ;
  $sCommand = "$sSAPWork/kill.sap";
  $sCommandArgs = "";

  ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 

  &logTrace ( 20046,
  "<%s> Stopping SAPWebAS71 with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20047,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          SAPWebAS71::ExecLocalCmd ( $sSAPAdmin, 
                                     $sEnvFile, 
                                     $sCommand, 
                                     $sCommandArgs, 
                                     \$iExitCode, 
                                     $iTimeout, 
                                     \@lsCmdout ) ) {
    &logTrace ( 20048,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  } 
  else {
    #------------------------------------
    # Define stop command and arguments..
    #------------------------------------
    $sCommand = "$sSAPWork/shutdown.sap";
    $sCommandArgs = "";
    ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 

    &logTrace ( 20049,
    "<%s> Stopping SAPWebAS71 with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );

    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20050,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
            SAPWebAS71::ExecLocalCmd ( $sSAPAdmin, 
                                       $sEnvFile, 
                                       $sCommand, 
                                       $sCommandArgs, 
                                       \$iExitCode, 
                                       $iTimeout, 
                                       \@lsCmdout ) ) {
      &logTrace ( 20051,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      #--------------------
      # Get time bearings..
      #--------------------
      $iSleepTime = 10; # Secs..

      @liPids = $bSapStart ? shift @liRunningProcs : @liRunningProcs; # Global..

      if( $bSapStart && defined $iCCMSX ) {
        push ( @liPids, $iCCMSX );
      }

      if( $bSapStart && defined $iCCMSR ) {
        push ( @liPids, $iCCMSR );
      }

      #-----------------------------------------------
      # Attempt to request sapstart to stop instance..
      #-----------------------------------------------
      if ( scalar @liPids ) {
        &logInfo ( 20052, "<%s> Sending SIGINT to PIDs [%s]",
        $sMyName, join ( ' ', @liPids ) );

        #-----------------------------------------
        # Give SAP chance to shutdown gracefully..
        #-----------------------------------------
        if ( kill ( 2, @liPids ) == scalar @liPids ) {
          &logInfo ( 20053,
          "<%s> Successfully sent SIGINT to PIDs [%s]",
          $sMyName, join ( ' ', @liPids ) );
        }
        else {
          &logWarn ( 20054,
          "<%s> Failed to send SIGINT to PID [%s]",
          $sMyName, join ( ' ', @liPids ) );
        }
      }
    }
  }

  #--------------------
  # Delay entry point..
  #--------------------
  if ( $iReturnCode ) { # Success..
    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20055,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      SAPWebAS71::GetProcState ( \@liPids,
                                 $sInstProfile,
                                 $sInstType,
                                 $sProcMon,
                                 $sSAPAdmin,
                                 0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        #-----------------------------------------------
        # Check if the last process is not sapstartsrv..
        #-----------------------------------------------
        if ( 1 == scalar @liPids   && # Last remaining..
             defined $iSapStartSrv && #
             $iSapStartSrv == $liPids [0] ) {
          last;
        }

        &logTrace ( 20056,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout - $iSleepTime );
      }
      else {
        &logInfo ( 20057,
        "<%s> Confirmed shutdown of SAPWebAS71 instance",
        $sMyName );
        last; # Get out..
      }
    }
  }

  #-----------------------------------------
  # Whack any persisting SAPWebAS71 processes..
  #-----------------------------------------
  &logTrace ( 20058,
  "<%s> Killing SAPWebAS71 processes for this instance", $sMyName );

  if ( SAPWebAS71::KillSAPWebAS71Processes ( $sInstProfile,
                                             $sInstType,
                                             $sProcMon,
                                             $sSAPAdmin ) ) {
    &logTrace ( 20059,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20060,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }

  my ($sKillFile, $sShutdownFile, $sCollLockFile, $sSendLockFile, @lsFiles ) = ();

  #------------------
  # Remove kill.sap..
  #------------------
  $sKillFile = "/usr/sap/$sSAPSID/$sInstName/work/kill.sap";

  if ( -e $sKillFile ) {
    &logTrace ( 20061,
    "<%s> Located file [%s] for removal", $sMyName, $sKillFile );
    push ( @lsFiles, $sKillFile );
  }

  #----------------------
  # Remove shutdown.sap..
  #----------------------
  $sShutdownFile = "/usr/sap/$sSAPSID/$sInstName/work/shutdown.sap";

  if ( -e $sShutdownFile ) {
    &logTrace ( 20062,
    "<%s> Located file [%s] for removal", $sMyName, $sShutdownFile );
    push ( @lsFiles, $sShutdownFile );
  }

  #-----------------------------
  # Remove collector lock file..
  #-----------------------------
  $sCollLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgcpid";

  if ( -e $sCollLockFile ) {
    &logTrace ( 20063,
    "<%s> Located file [%s] for removal", $sMyName, $sCollLockFile );
    push ( @lsFiles, $sCollLockFile );
  }

  #--------------------------
  # Remove sender lock file..
  #--------------------------
  $sSendLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgspid";

  if ( -e $sSendLockFile ) {
    &logTrace ( 20064,
    "<%s> Located file [%s] for removal", $sMyName, $sSendLockFile );
    push ( @lsFiles, $sSendLockFile );
  }

  #-------------------
  # Remove the files..
  #-------------------
  if ( scalar @lsFiles && ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
    &logError ( 20065, "<%s> Failed to remove stale file(s) [%s]",
    $sMyName, join ( ' ', @lsFiles ) );
    $iReturnCode = 0; # Failure..
  }
  
  #-----------------------------------
  # Check if it's safe to clean IPCs..
  #-----------------------------------
  if ( $sInstType =~ /^(appserv|enqrep)$/io ) {
    #----------------------------------------
    # Define cleanipc command and arguments..
    #----------------------------------------
    my $iInstID = ();
    ($iInstID ) = $sInstName =~ /(\d{2})$/o;
    $sCommand = "$sSAPHome/cleanipc";
    $sCommandArgs = "$iInstID remove";
    ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate..

    &logTrace ( 20066,
    "<%s> Stopping SAPWebAS71 with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );

    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20067,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
            SAPWebAS71::ExecLocalCmd ( $sSAPAdmin,
                                       $sEnvFile,
                                       $sCommand,
                                       $sCommandArgs,
                                       \$iExitCode,
                                       $iTimeout,
                                       \@lsCmdout ) ) {
      &logTrace ( 20068,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }
  else {
    &logTrace ( 20069,
    "<%s> Skipping cleanipc for InstType [%s].
    Possibility of destroying active lock table", $sMyName, $sInstType );
  }

  #----------------------------------
  # Initiate SAPWebAS71 shutdown log..
  #----------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sMessage = "Shutdown completed for [$sResourceName]";
  if ( ! ( $sInstType =~ /smdagent/io ) ) {
    if ( SAPWebAS71::UpdateSAPWebAS71Log ( $sInstName,
                                           $sInstProfile,
                                           $sInstType,
                                           $sSAPAdmin,
                                           $sSAPSID,
                                           $sMessage ) ) {
      &logTrace ( 20070,
      "<%s> Updated log successfully", $sMyName );
    }
    else {
      &logError ( 20071,
      "<%s> Could not log successfully", $sMyName );
    }
  }

  SAPWebAS71::DeletePidFile ( $sResourceName );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String EnvFile,
                        IN String InstProfile,
                        IN String InstType,
                        IN String ProcMon,
                        IN String SAPAdmin, 
                        IN String SAPSID, 
                        IN String SAPHome, 
                        IN String InstName );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $sSAPSID                   SAP System ID
 $sSAPHome                  SAP Instance Home
 $sInstName                 SAP Instance Name
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sEnvFile,
       $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin,
       $sSAPSID,
       $sSAPHome,
       $sInstName ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20072, "Subroutine <%s> called with:
  CleanReason              [%s]
  EnvFile                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  SAPHome                  [%s]
  InstName                 [%s]",

  $sMyName,  $iCleanReason,  $sEnvFile,  $sInstProfile, $sInstType,
  $sProcMon, $sSAPAdmin, $sSAPSID, $sSAPHome, $sInstName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPWebAS71::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsCmdout, @liDummyPids, @liPids ) = ();

  my ($sKillFile, $sShutdownFile, $sCollLockFile, $sSendLockFile, @lsFiles ) = ();
  
  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #----------------------------
    # Get SAP PIDs
    #----------------------------
    SAPWebAS71::GetProcState ( \@liDummyPids,
                               $sInstProfile,
                               $sInstType,
                               $sProcMon,
                               $sSAPAdmin,
                               0 ); # Use RFLM..

    @liPids = $bSapStart ? shift @liDummyPids: @liDummyPids; # Local..

    if( $bSapStart && defined $iCCMSX ) {
      push ( @liPids, $iCCMSX );
    }

    if( $bSapStart && defined $iCCMSR ) {
      push ( @liPids, $iCCMSR );
    }

    #-----------------------------------------------
    # Attempt to request sapstart to stop instance..
    #-----------------------------------------------
    if ( scalar @liPids ) {
      &logInfo ( 20073, "<%s> Sending SIGINT to PIDs [%s]",
      $sMyName, join ( ' ', @liPids ) );

      #-----------------------------------------
      # Give SAP chance to shutdown gracefully..
      #-----------------------------------------
      if ( kill ( 2, @liPids ) == scalar @liPids ) {
        &logInfo ( 20074,
        "<%s> Successfully sent SIGINT to PIDs [%s]",
        $sMyName, join ( ' ', @liPids ) );
      }
      else {
        &logWarn ( 20075,
        "<%s> Failed to send SIGINT to PID [%s]",
        $sMyName, join ( ' ', @liPids ) );
      }
    }
    else {
      &logError ( 20076,
      "<%s> Proceeding with SIGKILL for SAP Instance [%s]",
      $sMyName, $sInstName );
    }
  } 

  #-----------------------------------------
  # Whack any persisting SAPWebAS71 processes..
  #-----------------------------------------
  &logTrace ( 20077,
  "<%s> Killing SAPWebAS71 processes for this instance", $sMyName );

  if ( SAPWebAS71::KillSAPWebAS71Processes ( $sInstProfile,
                                             $sInstType,
                                             $sProcMon,
                                             $sSAPAdmin ) ) {
    &logTrace ( 20078,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20079,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  my $iInstID = ();

  #-----------------------------------
  # Check if it's safe to clean IPCs..
  #-----------------------------------
  if ( $sInstType =~ /^(appserv|enqrep)$/io ) {
    ($iInstID ) = $sInstName =~ /(\d{2})$/o;

    $sCommand = "$sSAPHome/cleanipc";
    $sCommandArgs = "$iInstID remove";

    #------------------------
    # Run cleanipc ..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20080,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( SAPWebAS71::ExecLocalCmd ( $sSAPAdmin,
                                       $sEnvFile,
                                       $sCommand,
                                       $sCommandArgs,
                                       \$iExitCode,
                                       $iTimeout,
                                       \@lsCmdout ) ) {
      &logTrace ( 20081,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      &logError ( 20082, "<%s> Failed to execute cleanipc.", $sMyName );
      $iReturnCode = 0; # Failure..
    }
  }
  else {
    &logTrace ( 20083,
    "<%s> Skipping cleanipc for InstType [%s].", $sMyName, $sInstType );
  }

  #------------------
  # Remove kill.sap..
  #------------------
  $sKillFile = "/usr/sap/$sSAPSID/$sInstName/work/kill.sap";

  if ( -e $sKillFile ) {
    &logTrace ( 20084,
    "<%s> Located file [%s] for removal", $sMyName, $sKillFile );
    push ( @lsFiles, $sKillFile );
  }

  #----------------------
  # Remove shutdown.sap..
  #----------------------
  $sShutdownFile = "/usr/sap/$sSAPSID/$sInstName/work/shutdown.sap";

  if ( -e $sShutdownFile ) {
    &logTrace ( 20085,
    "<%s> Located file [%s] for removal", $sMyName, $sShutdownFile );
    push ( @lsFiles, $sShutdownFile );
  }

  #-----------------------------
  # Remove collector lock file..
  #-----------------------------
  $sCollLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgcpid";

  if ( -e $sCollLockFile ) {
    &logTrace ( 20086,
    "<%s> Located file [%s] for removal", $sMyName, $sCollLockFile );
    push ( @lsFiles, $sCollLockFile );
  }

  #--------------------------
  # Remove sender lock file..
  #--------------------------
  $sSendLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgspid";

  if ( -e $sSendLockFile ) {
    &logTrace ( 20087,
    "<%s> Located file [%s] for removal", $sMyName, $sSendLockFile );
    push ( @lsFiles, $sSendLockFile );
  }

  #-------------------
  # Remove the files..
  #-------------------
  if ( scalar @lsFiles && ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
    &logError ( 20088, "<%s> Failed to remove stale file(s) [%s]",
    $sMyName, join ( ' ', @lsFiles ) );
    $iReturnCode = 0; # Failure..
  }

  #----------------------------------
  # Initiate SAPWebAS71 cleanup log..
  #----------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sMessage = "Clean called for [$sResourceName] [$sCleanReason]";

  if ( ! ( $sInstType =~ /smdagent/io ) ) {
    if ( SAPWebAS71::UpdateSAPWebAS71Log ( $sInstName,
                                           $sInstProfile,
                                           $sInstType,
                                           $sSAPAdmin,
                                           $sSAPSID,
                                           $sMessage ) ) {
      &logTrace ( 20089,
      "<%s> Updated log successfully", $sMyName );
    }
    else {
      &logError ( 20090,
      "<%s> Could not log successfully", $sMyName );
    }
  }

  SAPWebAS71::DeletePidFile ( $sResourceName );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String InstProfile,
                                   IN String InstType,
                                   IN String ProcMon,
                                   IN String SAPAdmin );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20091, "Subroutine <%s> called:
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]",

  $sMyName, $sInstProfile, $sInstType, $sProcMon, $sSAPAdmin );
  

  my @liDummyPids = ();
  my $iReturnCode = SAPWebAS71::GetProcState ( \@liDummyPids,
                                               $sInstProfile,
                                               $sInstType,
                                               $sProcMon,
                                               $sSAPAdmin,
                                               1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20092,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20093,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

                                                                                
  &logTrace ( 20094,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String InstProfile,
                               IN String InstType,
                               IN String ProcMon,
                               IN String SAPAdmin,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20095, "Subroutine <%s> called with:
  OnlinePids               [%s]
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  UseFFLM                  [%s]",

  $sMyName, $rliOnlinePids, $sInstProfile, $sInstType, $sProcMon,
  $sSAPAdmin, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp, $iSapstartPid,
       @hProcListHash, @lsUpdates, @lsValues, $iDwMaster, $iJcFlag,
       %hiProcPids, %hsSrvrDetails, %hsCopy, $sEntryPointName,
       $iSleepTime, $sInstName, $sSAPSID, @liDummyPids, %hsProcMonCopy, 
       $iJstartFlag, $iJstartmFlag, $sSAPHost, @lsWords, @lsOthers ) = ();

  my ( $sSwitch, $bSapstartFlag, $iPosn, $bAddIn, %hiDwPids, $iParent, $i ) = ();

  @$rliOnlinePids = ();
  
  #--------------------------
  # Other processes to kill..
  #--------------------------
  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);

  @lsOthers = ( "pf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}" );

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  my $sArch = 0;
  my %hsArch = ( 1 => 'Java',
                 2 => 'ABAP',
                 3 => 'AddIn' );

  my @lsProcMon = $sProcMon =~ /([^\s]+)/g;
  my %hsProcs = map { $_ => 1 } @lsProcMon;

  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  if ( $hsProcs { jc } ) {
    $sArch = 1;
  }
  elsif ( $hsProcs { jstart } && $hsProcs { dw } ) {
    $sArch = 3;
    ++ $bAddIn;
  }
  elsif ( $hsProcs { JSTARTM } ) { # Recursive call..
    $sArch = 3;
    ++ $bAddIn;
  }
  elsif ( $hsProcs { dw } ) {
    $sArch = 2;
  }

=pod
  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  if( 1 == grep { /^(jc)$/o } @lsProcMon ) {
    $sArch = 1;
  }
  elsif ( 2 == grep { /^(jstart|dw)$/o } @lsProcMon ) {
    $sArch = 3;
    ++ $bAddIn;
  }
  elsif ( 1 == grep { /^(dw)$/o } @lsProcMon ) {
    $sArch = 2;
  }
=cut

  if ( $sArch ) {
    &logTrace ( 20096,
    "<%s> SAP Instance Architecture is [%s]",
    $sMyName, $hsArch { $sArch } );
  }
  else {
    &logTrace ( 20097,
    "<%s> Couldn't determine SAP Architecture",
    $sMyName );
  }

  #-----------------------------------------------
  # Insert sapstart, unless its a recursive call..
  #-----------------------------------------------
  if ( $sProcMon eq 'JSTART' ) { # It's Java ..
    $sProc = 'jstart';
  }
  elsif ( $sProcMon eq 'JSTARTM' ) { # It's AddIn ..
    $sProc = 'dw jstart'; # AddIn for recursive call..
  }
  else {
    $sProc =  "sapstart $sProcMon";
  }

  $iUid = getpwnam $sSAPAdmin;

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == $iUid ) &&      # User is the super-user..
       ( 2 == $iUseFFLM ) ) { # And called to kill-by-UID..
    &logTrace ( 20098, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sSAPAdmin, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( SAPWebAS71::GetSrvrDetails ( $sInstProfile, 
                                      $sProc, 
                                      \%hsSrvrDetails ) ) {

      delete $hsSrvrDetails { dw } if $sProcMon eq 'JSTARTM'; # Remove dw in recursive..

      &logTrace ( 20099,
      "<%s> Got details for monitoring SAPWebAS71 instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20100,
      "<%s> Failed to get details for monitoring SAPWebAS71 instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $sEntryPointName ne 'offline' && $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nSAPWebAS71 process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = SAPWebAS71::FFLMStatus ( $sPidFile,
                                             $sProc,
                                             $iUid,
                                             \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          if ( $sKey eq 'sapstart' ) { # Check sapstart found ..
            $iSapstartPid = $iPid; 
            unshift ( @$rliOnlinePids, $iPid ); # Insert at the front..
          }
          else {
            if ( $sKey eq 'dw' ) {  # Check dw master found for ABAP or AddIn ..
              $iDwMaster = $iPid; 
            }
            elsif ( $sKey eq 'jc' ) { # Check jc found for Java ..
              ++ $iJcFlag;
            }
            elsif ( $sKey eq 'jstartm' ) { # Check jstart master found for AddIn ..
              ++ $iJstartmFlag;
            }
            elsif ( $sKey eq 'jstart' ) {
              ++ $iJstartFlag;
            }

            &logTrace ( 20101, "<%s> [%s] found in FFLM [%s]", $sMyName, $sKey, $iPid);

            push ( @$rliOnlinePids, $iPid ); # Push..
          }
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20102,
      "<%s> SAPWebAS71 instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20103,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sSAPAdmin, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20104,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20105,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20106,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    my $iParent = 1; # Set to init..
    my $sOSVer = $^O eq 'solaris' ? ( uname ) [2] : '';

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20107, "<%s> Searching UID [%s] PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID }, $$rhlProc { COMMAND } );

      #--------------------------------
      # Handle Solaris 10 local zones..
      #--------------------------------
      if ( $sOSVer eq '5.10'                       && # Solaris 10..
           $$rhlProc { COMMAND } =~ m@/sbin/init$@ && # With init..
           ! kill ( 0, 1 ) ) {                        # In a non-global zone..
        $iParent = $$rhlProc { PPID };                # Get actual PID..
      }

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) = splice ( @lsValues, 0, 3 ) ) {
              if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
                if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..
                  #-----------------------
                  # Handle dw separately..
                  #-----------------------
                  $iFound = VCSDef_MonitorExit::ONLINE unless $sKey eq 'dw';

                  #-----------------------
                  # Handle special cases..
                  #-----------------------
                  $sSwitch = $sKey =~ /^(jc|jstart|dw|sapstart)$/o ? $sKey: 'default';

                  goto $sSwitch; {
                    sapstart:
                      $iPosn = 0;
                      $iSapstartPid = $iPid; # Store..
                    last;

                    jc:
                      $iPosn = 1;
                      ++ $iJcFlag; # Set flag..
                    last;

                    jstartm:
                      $iPosn = 2;
                      ++ $iJstartmFlag;
                    last;

                    jstart:
                      $iPosn = 3;
                      ++ $iJstartFlag;
                    last;
                    dw:
                      $iPosn = 4;
                      $hiDwPids { $iPid } = $$rhlProc { PPID }; # Store..
                    last;
 
                    default:
                      $iPosn = scalar @$rliOnlinePids;
                    last;
                  }

                  #--------------------------------
                  # Insert into appropriate place..
                  #--------------------------------
                  $iPosn = scalar @$rliOnlinePids if $iPosn > scalar @$rliOnlinePids;

                  splice ( @$rliOnlinePids, $iPosn, 0, $iPid ); # Insert..
                  $hiProcPids { $sKey } = $iPid;                # Store..
    
                  &logTrace ( 20108,
                  "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                  $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND }, $sProc );
                }
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        elsif ( $sProcMon !~ /JSTART/ ) { # Request for kill..
          &logTrace ( 20109,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }
      
      #---------------------------------
      # Handle other related processes..
      #---------------------------------
      if ( ( ! grep { /^$iPid$/ } @$rliOnlinePids ) &&         # Not caught earlier..
           $$rhlProc { UID } == $iUid               &&         # Owned by SAPAdmin
           map { $$rhlProc { COMMAND } =~ /$_/ } @lsOthers ) { # Was now found..
        push ( @$rliOnlinePids, $iPid );                       # Store..

        #-----------------------------------------
        # Populate dwindling sapstartsrv process..
        #-----------------------------------------
        if ( $$rhlProc { COMMAND } =~ /sapstartsrv/o ) {
          $iSapStartSrv = $iPid; # Populate global..
        }
        elsif ( $$rhlProc { COMMAND } =~ /cm\.sap/o ) { # Note sapccm4x process
          $iCCMSX = $iPid; # Populate global..
        }
        elsif ( $$rhlProc { COMMAND } =~ /cs\.sap/o ) { # Note sapccmsr process
          $iCCMSR = $iPid; # Populate global..
        }
        elsif ( $$rhlProc { COMMAND } =~ /sapccm4x/o ) {
          $iCCMSX = $iPid; # Populate global..
        }
        elsif ( $$rhlProc { COMMAND } =~ /sapccmsr/o ) { # Note sapccmsr process
          $iCCMSR = $iPid; # Populate global..
        }

        &logTrace ( 20110,
        "<%s> Matched process PID [%s] PPID [%s] CMD [%s]",
        $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
      }
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline ) = ();

  my @liDwPids = keys %hiDwPids;
  $sTemp = ''; # Ensure..

  foreach $sKey ( sort keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) = splice ( @lsValues, 0, 3 ) ) {
      #-------------------------------------
      # Set status for dw master process ..
      #-------------------------------------
      if ( $bAddIn && $sProcMon !~ /JSTARTM/ ) {
        &logTrace ( 20111, 
        "<%s> Key [%s] EntryPoint [%s] AddIn [%s] dw master [%s]",
        $sMyName, $sKey, $sEntryPointName, $bAddIn, $iDwMaster );
      }

      if ( $sKey eq 'dw' && scalar @liDwPids ) {
        #-------------------------------------
        # Check if we the dw master process ..
        #-------------------------------------
        if ( $iSapstartPid &&                    # sapstart exists
             ( ( $iDwMaster ) =                  # Get the dw master 
             grep {                              # By finding
               $iSapstartPid == $hiDwPids { $_ } # sapstart as it's parent
             } @liDwPids ) ) {                   # From all dw processes
          $hiProcPids { $sKey } = $iDwMaster;    # Store..
          $iFound = VCSDef_MonitorExit::ONLINE;  # And mark online..
          &logTrace ( 20112,
          "<%s> Found master dw process PID [%s] PPID [%s]",
          $sMyName, $iDwMaster, $iSapstartPid );
        }
        elsif ( ( ( $iDwMaster ) =               # A dw process..
                grep {                           # Found..
                  $iParent == $hiDwPids { $_ } &&# With init as parent..
                  ++ $i                          # Mark it..
                } @liDwPids ) && 1 == $i ) {     # If only one such exists..
          $hiProcPids { $sKey } = $iDwMaster;    # Store..
          $iFound = VCSDef_MonitorExit::ONLINE;  # And mark online..
          &logWarn ( 20113,
          "<%s> Missing sapstart process.
          Found master dw process PID [%s] PPID [%s]",
          $sMyName, $iDwMaster, $hiDwPids { $iDwMaster } );
        }
        else {
          &logTrace ( 20114,
          "<%s> Missing dw master. Marking status as offline", $sMyName );

          $iFound = VCSDef_MonitorExit::OFFLINE; # Ensure its offline..

          #------------------------
          # Delete PID, if stored..
          #------------------------
          if ( defined $hiProcPids { $sKey } ) {
            delete $hiProcPids { $sKey };
          }
        }
      }

      #--------------------------
      # Mark status of sapstart..
      #--------------------------
      if ( $sKey eq 'sapstart' && $iSapstartPid ) {
        ++ $bSapstartFlag; # Mark it found..
        ++ $bSapStart;     # Mark global..
      }

      #------------------------------------
      # If we are AddIn, validate jstart ..
      #------------------------------------
      if ( $sKey eq 'jstartm'            && # Looking for jstart..
           $sEntryPointName eq 'monitor' && # in monitor..
           $bAddIn                       && 
           $iDwMaster                   ) { # dw and jstart declared..

        #-----------------------------------
        # Calculate sleep time dynamically..
        #-----------------------------------
        my $iTillWhen = $^T + SAPWebAS71::GetEngineTimeout ( 1 );
        $iSleepTime = ( $iTillWhen - time ) / 10;
        my $iTimeRemain = undef;

        while ( ! ( $iJstartmFlag && $iJstartFlag ) ) { # jstart master missing..
           
          unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
            &logTrace ( 20115,
            "<%s> Insufficient time [%s] sec to retry. Exiting ..", $sMyName, $iTimeRemain );

            last;
          }

          &logWarn ( 20116,
          "<%s> Missing 'jstart' master, though 'dw' master [%s] exists.
          Retrying in [%s] sec", $sMyName, $iDwMaster, $iSleepTime );

          sleep $iSleepTime; # Take a break..

          #-------------------------------
          # Call ourselves with 'JSTARTM'..
          #-------------------------------
          $iFound = SAPWebAS71::GetProcState ( \@liDummyPids,
                                               $sInstProfile,
                                               $sInstType,
                                               'JSTARTM',      # In CAPS..
                                               $sSAPAdmin,
                                               0 );            # Regular FLM..

          #--------------------------------------------------------------------
          # If we returned online, we know we found jstart master and at least 
          # one jstart server process. PID now is in first position..
          #--------------------------------------------------------------------
          if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
            &logTrace ( 20117,
            "<%s> Found jstart master process after recursive call",
            $sMyName );

            #-----------------------------
            # Update jstartm and jstart ..
            #-----------------------------
            ( $hiProcPids { jstartm }, $hiProcPids { jstart } ) = (); # Ensure undefs..so that PID 
                                                                      # files are not created again for FFLM..

            ++ $iJstartmFlag; # Set flag..
            ${ $hsCopy { jstartm } } [-1] = $iFound; # Mark online..

            ++ $iJstartFlag; # Set flag..
            ${ $hsCopy { jstart } } [-1] = $iFound; # Mark online..
          }
          else {
            &logError ( 20118, "<%s> Missing jstart child process after recursive call. Retrying ..", $sMyName );

            ${ $hsCopy { jstartm } } [-1] = VCSDef_MonitorExit::ONLINE;
            ${ $hsCopy { jstart } } [-1] = VCSDef_MonitorExit::OFFLINE;

            $iFound = VCSDef_MonitorExit::OFFLINE;
          }
        }
      }

      #--------------------------------------------------
      # Check if 'jstart' is missing and 'jc' is alive ..
      #--------------------------------------------------
      if ( $sEntryPointName eq 'monitor' && # In monitor..
           $sKey eq 'jstart'             && # Jstart,
           $iJcFlag                     ) { # when jc, alive..

        #-----------------------------------
        # Calculate sleep time dynamically..
        #-----------------------------------
        my $iTillWhen = $^T + SAPWebAS71::GetEngineTimeout ( 1 );
        $iSleepTime = ( $iTillWhen - time ) / 10;

        my $iTimeRemain = undef;

        while ( ! $iJstartFlag ) { # jstart missing..

          unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
            &logTrace ( 20119,
            "<%s> Insufficient time [%s] sec to retry. Exiting..", $sMyName, $iTimeRemain );

            last;
          }

          &logWarn ( 20120,
          "<%s> Missing process [%s], though 'jc' exists. Retrying in [%s] sec",
          $sMyName, $sKey, $iSleepTime );

          sleep $iSleepTime; # Take a break..

          #--------------------------------------
          # Call ourselves again with 'jstart' ..
          #--------------------------------------
          $iFound = SAPWebAS71::GetProcState ( \@liDummyPids,
                                               $sInstProfile,
                                               $sInstType,
                                               'JSTART',
                                               $sSAPAdmin,
                                               0 ); # Regular FLM..
          #--------
          # Store..
          #--------
          if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
            $hiProcPids { jstart } = (); 
            ${ $hsCopy { jstart } } [-1] = $iFound; # Mark online..
            ++ $iJstartFlag;

            &logTrace ( 20121,
            "<%s> Found jstart process after recursive call.", $sMyName );
          }
          else {
            &logError ( 20122, "<%s> Missing jstart process after recursive call. Retrying ..", $sMyName );
            $iFound = VCSDef_MonitorExit::OFFLINE; # Ignore partial..
          }
        }
      }

      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM && $iUseFFLM != 2               &&      # In RFLM..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20123,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( SAPWebAS71::CreatePidFile ( $sPidFile, $hiProcPids { $sKey } ) ) {
          &logTrace ( 20124,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20125,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound && $sKey ne 'sapstart' ) {
        if ( $bAddIn && $sKey eq 'jstart' && $sProcMon !~ /JSTART/ ) {
          $iFound = 'XXX';
        }
        else {
          $iOffline = $iFound;
        }
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOnline && $iOffline ? VCSDef_MonitorExit::PARTIAL_ONLINE:
                 $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #$sTemp =~  s/XXX/VCSDef_MonitorExit::ONLINE == $iReturnCode ? 110:100/e;

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    $sTemp =~ s/XXX/110/;
  }
  else {
    $sTemp =~ s/XXX/100/;
  }

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20126,
    "<%s> Found [%s] PIDs [%s] for this instance of SAPWebAS71.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #---------------------
  # Report on sapstart..
  #---------------------
  if ( ! $bSapstartFlag                    &&          # Missing sapstart..
       $sEntryPointName eq 'monitor'       &&          # In monitor..
       $sProcMon !~ /^(JSTART|JSTARTM)$/o  &&          # Not in recursive call..
       VCSDef_MonitorExit::OFFLINE != $iReturnCode ) { # Not offline..

    &logWarn ( 20127,
    "<%s> Missing sapstart process in first level monitor",
    $sMyName );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20128,
    "<%s> SAPWebAS71 process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @$rliOnlinePids;

  &logTrace ( 20129,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String EnvFile,
                          IN String ProcMon,
                          IN String InstProfile,
                          IN String InstType,
                          IN String MonitorProgram,
                          IN String SAPAdmin,
                          IN Integer SecondLevelMonitor );

 Description: Performs the in-depth monitoring

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sProcMon                  List of Processes to be monitored 
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sMonitorProgram           Full path to external monitor program
 $sSAPAdmin                 SAP System Administrator
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sProcMon,
       $sInstProfile,
       $sInstType,
       $sMonitorProgram,
       $sSAPAdmin,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20130, "Subroutine <%s> called with:
  EnvFile                  [%s]
  ProcMon                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  MonitorProgram           [%s]
  SAPAdmin                 [%s]
  SecondLevelMonitor       [%s]",

  $sMyName, $sEnvFile, $sProcMon, $sInstProfile, $sInstType, 
  $sMonitorProgram, $sSAPAdmin, $iSecondLevelMonitor );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPWebAS71::DelayedAttributeCheck ( $sEnvFile,
                                             $sInstProfile,
                                             $sSAPAdmin ) ) ) {
      &logTrace ( 20131,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20132,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20133,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + SAPWebAS71::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPWebAS71::SecondLevelMonitor ( $sEnvFile,
                                          $sProcMon,
                                          $sInstProfile,
                                          $sInstType,
                                          $sSAPAdmin,
                                          $iTillWhen ) ) ) {

      &logTrace ( 20134,
      "<%s> SAPWebAS71 instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20135,
      "<%s> SAPWebAS71 instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + SAPWebAS71::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         SAPWebAS71::ExternalMonitor ( $sSAPAdmin,
                                       $sEnvFile,
                                       $sExtCmd,
                                       $sExtCmdArgs,
                                       $iTillWhen ) ) ) {
      &logTrace ( 20136,
      "<%s> SAPWebAS71 instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20137,
      "<%s> SAPWebAS71 instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String EnvFile,
                                        IN String InstProfile,
                                        IN String SAPAdmin );

 Description : Just In Time Validation

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sSAPAdmin                 SAP System Administrator
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstProfile,
       $sSAPAdmin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20138, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstProfile              [%s]
  SAPAdmin                 [%s]",

  $sMyName, $sEnvFile, $sInstProfile, $sSAPAdmin );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( EnvFile InstProfile SAPAdmin );
  my @lsReqdArgs = qw ( EnvFile InstProfile SAPAdmin );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20139,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #-------------------------------
  # Validate parameter SAPAdmin..
  #-------------------------------
  if ( ! defined getpwnam $sSAPAdmin ) {
    &logError ( 20140, "<%s> SAPAdmin [%s] is not a valid UNIX login",
    $sMyName, $sSAPAdmin );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( ! -e $sEnvFile ) {
    &logError ( 20141, "<%s> EnvFile [%s] file does not exist",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #----------------------------------
  # Validate parameter InstProfile..
  #----------------------------------
  if ( ! -e $sInstProfile ) {
    &logError ( 20142, "<%s> InstProfile [%s] file does not exist",
    $sMyName, $sInstProfile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String EnvFile,
                                     IN String ProcMon,
                                     IN String InstProfile,
                                     IN String InstType,
                                     IN String SAPAdmin,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sProcMon                  List of processes to be monitored
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sSAPAdmin                 SAP System Administrator
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sProcMon,
       $sInstProfile,
       $sInstType,
       $sSAPAdmin,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20143, "Subroutine <%s> called with:
  EnvFile                  [%s]
  ProcMon                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  SAPAdmin                 [%s]
  TillWhen                 [%s]",

  $sMyName, $sEnvFile, $sProcMon, $sInstProfile, $sInstType,
  $sSAPAdmin, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();

  my ($sSAPSID, $sInstName, $iInstID, $sSAPHost, $sSAPHome) = ();
  my (@lsWords) = ();

  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;
  $sSAPHome = "/usr/sap/$sSAPSID/$sInstName/exe";

  my ( $sSwitch, $iRetCode ) = ();

  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  my $sArch = 0;
  my %hsArch = ( 1 => 'Java',
                 2 => 'ABAP',
                 3 => 'AddIn' );

  my @lsProcMon = $sProcMon =~ /([^\s]+)/g;

  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  if( 1 == grep { /^(jc)$/o } @lsProcMon ) {
    $sArch = 1;
  }
  elsif ( 2 == grep { /^(jstart|dw)$/o } @lsProcMon ) {
    $sArch = 3;
  }
  elsif ( 1 == grep { /^(dw)$/o } @lsProcMon ) {
    $sArch = 2;
  }

  if ( $hsArch { $sArch } ) {
    $sSwitch = $hsArch { $sArch };
    &logTrace ( 20144,
    "<%s> Identified SAP Architecture [%s] from ProcMon [%s]",
    $sMyName, $sSwitch, $sProcMon );
  }
  else {
    &logTrace ( 20145,
    "<%s> Could not identify SAP Architecture from ProcMon [%s]",
    $sMyName, $sProcMon );
    $sSwitch = 'DEFAULT'; # Mark it default..
  }

  #-----------------------
  # Do the correct stuff..
  #-----------------------
  goto $sSwitch; {

    Java:

      if ( $sInstType =~ /^(appserv|smdagent)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCommand = "$sSAPHome/sapcontrol";
        $sCommandArgs = " -nr $iInstID -function J2EEGetProcessList";
        $iTimeout = $iTillWhen - time;

        $iReturnCode = SAPWebAS71::ExecJSMon ( $sSAPAdmin,
                                               $sEnvFile,
                                               $sCommand,
                                               $sCommandArgs,
                                               $iTimeout );
      }
      else {
        &logError ( 20146,
        "<%s> Encountered unsupported InstType [%s] for SAP Arch [%s]",
        $sMyName, $sInstType, $sSwitch );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

    last;

    ABAP:

      if ( $sInstType =~ /^(appserv)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCommand = "$sSAPHome/sapinfo";
        $sCommandArgs = "ashost=$sSAPHost sysnr=$iInstID";
        $iTimeout = $iTillWhen - time;

        $iReturnCode = SAPWebAS71::ExecSAPInfo ( $sSAPAdmin, # User context..
                                                 $sEnvFile,  # Use envfile..
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 $iTimeout );
      }
      else {
        &logError ( 20147,
        "<%s> Encountered unsupported InstType [%s] for SAP Arch [%s]",
        $sMyName, $sInstType, $sSwitch );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

    last;

    AddIn:

      if ( $sInstType =~ /^(appserv)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCommand = "$sSAPHome/sapinfo";
        $sCommandArgs = "ashost=$sSAPHost sysnr=$iInstID";
        $iTimeout = $iTillWhen - time;

        $iReturnCode = SAPWebAS71::ExecSAPInfo ( $sSAPAdmin, # User context..
                                                 $sEnvFile,  # Use envfile..
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 $iTimeout );

        #---------------------
        # If offline, return..
        #---------------------
        if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ||
             VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
          return $iReturnCode;
        }
      }
      if ( $sInstType =~ /^(appserv|smdagent)$/io ) {  
        #--------------------------
        # Re-set command and args..
        #--------------------------

        $sCommand = "$sSAPHome/sapcontrol";
        $sCommandArgs = " -nr $iInstID -function J2EEGetProcessList";
        $iTimeout = $iTillWhen - time;

        $iRetCode = SAPWebAS71::ExecJSMon ( $sSAPAdmin,
                                            $sEnvFile,
                                            $sCommand,
                                            $sCommandArgs,
                                            $iTimeout );
        #-------------------------------------------
        # Update, status only when it's not online..
        #-------------------------------------------
        unless ( VCSDef_MonitorExit::ONLINE == $iRetCode ) {
          $iReturnCode = $iRetCode;
        }
      }
      else {
        &logError ( 20148,
        "<%s> Encountered unsupported InstType [%s] for SAP Arch [%s]",
        $sMyName, $sInstType, $sSwitch );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

    last;

    DEFAULT:

      if ( $sInstType !~ /^(appserv)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCommand = "$sSAPHome/ensmon";
        $sCommandArgs = "pf=$sInstProfile 2";
        $iTimeout = $iTillWhen - time;

        $iReturnCode = SAPWebAS71::ExecENSMon ( $sSAPAdmin,
                                                $sEnvFile,
                                                $sCommand,
                                                $sCommandArgs,
                                                $sInstType,
                                                $iTimeout );
      }
      else {
        &logError ( 20149,
        "<%s> Encountered unsupported InstType [%s]. Has to be one of
        [enqueue enqrep]",
        $sMyName, $sInstType );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

    last;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String SAPAdmin,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sSAPAdmin                 SAP System Administrator
 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20150, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20151,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20152,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20153,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20154,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20155,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  SAPWebAS71::ExecLocalCmd ( $sSAPAdmin, 
                             $sEnvFile, 
                             $sCmd, 
                             $sCmdArgs, 
                             \$iExitCode, 
                             $iTimeout, 
                             \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20156,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20157,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20158, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20159, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20160,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20161,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20162,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool DeletePidFile ( IN String ResourceName );


 Description : Deletes PID files for the VCS resource.

 Parameters :

 $sResourceName ResourceName for which the PID files needs to be deleted.

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub DeletePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sResourceName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20329, "Subroutine <%s> called with:
  ResourceName        [%s]",

  $sMyName, $sResourceName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  my ( $sPidFileDir, $iRet ) = ();
  #-------------------
  # Delete PID files..
  #-------------------
  &logTrace ( 20330,
  "<%s> Deleting PID files for VCS resource [%s]",
  $sMyName, $sResourceName );

  $sPidFileDir = "/tmp/.VRTSSAPWebAS71/" . $sResourceName . "/PID" ;

  if ( -d $sPidFileDir ) {
    $iRet = system("/bin/rm -rf $sPidFileDir") ;

    if ( $iRet == 0 ) {
      &logTrace ( 20331,
      "<%s> Successfully deleted PID files in the directory [%s].",
      $sMyName, $sPidFileDir );
      $iReturnCode = 0;
    }
    else {
      &logWarn ( 20332,
      "<%s> Failed to delete PID files in directory [%s]. Remove the files manually.",
      $sMyName, $sPidFileDir );
      $iReturnCode = 1;
    }
  }
  else {
    &logTrace ( 20333,
    "<%s> PID files for VCS resource [%s] doesn't exist.",
    $sMyName, $sResourceName );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20163, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20164, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20165, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20166, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20167,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20168,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20169,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( SAPWebAS71::GetPidDetails ( $iPid,
                                     \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20170,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20171,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20172,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20173,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20174, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillSAPWebAS71Processes ( IN String InstProfile,
                                           IN String InstType,
                                           IN String ProcMon,
                                           IN String SAPAdmin );

 Description  : Kills processes for this instance of SAPWebAS71
 
 Parameters :

 $sInstProfile            Instance profile file for the SAP Instance
 $sInstType               Instance type for the SAP Instance
 $sProcMon                List of processes to be monitored
 $sSAPAdmin               SAP System Administrator
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillSAPWebAS71Processes {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20175, "Subroutine <%s> called with:
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]",

  $sMyName, $sInstProfile, $sInstType, $sProcMon, $sSAPAdmin );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'SAPWebAS71::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    SAPWebAS71::GetProcState ( \@liGoners,
                               $sInstProfile,
                               $sInstType,
                               $sProcMon,
                               $sSAPAdmin,
                               0 ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20176,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20177,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20178,
    "<%s> Could not locate any process this instance of SAPWebAS71",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20179, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20180,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20181,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20182,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20183,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20184,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20185,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20186,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20187,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20188, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20189,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  goto $sSwitch; {
    solaris:

      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;

    hpux:

      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;

    aix:

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;


    linux:

      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;
  }

  &logTrace ( 20190, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20191,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20192,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String InstProfile,
                                  IN String ProcMon,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sInstProfile              Instance profile file for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sProcMon,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20193, "Subroutine <%s> called with:
  InstProfile              [%s]
  ProcMon                  [%s]
  SrvrDetails              [%s]",

  $sMyName, $sInstProfile, $sProcMon, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $sTmpProc, $sSAPSID, $sInstName,
       $iInstID, $sSAPHost, $sSAPHome, @lsWords ) = ();

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20194, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20195, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  #----------------------
  # Decide architecture..
  #----------------------
  ( @lsWords ) = split( /\//, $sInstProfile );
  ( $sSAPSID, $sInstName, $sSAPHost ) = split ( /_/, $lsWords [-1] );
  ( $iInstID ) = $sInstName =~ /(\d{2})$/o;
  $sSAPHome = "/usr/sap/$sSAPSID/$sInstName/exe";

  my $sArch = 0;
  my %hsArch = ( 1 => 'Java',
                 2 => 'ABAP',
                 3 => 'AddIn' );

  my @lsProcMon = $sProcMon =~ /([^\s]+)/g;
  my %hsProcs = map { $_ => 1 } @lsProcMon;

  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  if ( $hsProcs { jc } ) {
    $sArch = 1;
  }
  elsif ( $hsProcs { jstart } && $hsProcs { dw } ) {
    $sArch = 3;
  }
  elsif ( $hsProcs { dw } ) {
    $sArch = 2;
  }

=pod
  if( 1 == grep { /^(jc)$/o } @lsProcMon ) {
    $sArch = 1;
  }
  elsif ( 2 == grep { /^(jstart|dw)$/o } @lsProcMon ) {
    $sArch = 3;
  }
  elsif ( 1 == grep { /^(dw)$/o } @lsProcMon ) {
    $sArch = 2;
  }
=cut

  if ( $sArch ) {
    &logTrace ( 20196,
    "<%s> SAP Instance Architecture is [%s]",
    $sMyName, $hsArch { $sArch } );
  }
  else {
    &logTrace ( 20197,
    "<%s> Couldn't determine SAP Architecture",
    $sMyName );
  }

  if ( 1 == $sArch ) { # jc located..
    push @lsProcMon, 'jstart';
  }
  elsif ( 3 == $sArch ) {  # jstart located..
    push @lsProcMon, 'jstartm';
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsProcMon ) {
    $sKey = "srchstr_${sProc}";
    $sPidFile = "$sPIDDir/pidfile.$sKey";

    if ( $sProc eq 'sapstart' ) {
      $sTmpProc = qr|$sProc\spf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}|;
    }
    elsif ( $sProc eq 'jstart' ) { # Child.. Atre 27 Mar 2009
      $sTmpProc = qr|$sProc\s.*\spf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}.*-hostvm|;
    }
    elsif ( $sProc eq 'jstartm' ) { # Parent.. Atre 27 Mar 2009
      $sTmpProc = qr|jstart\spf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}|;
    }
    elsif( $sProc eq 'gwrd' || $sProc eq 'icman' ){
      $sTmpProc = qr|$sProc\s.*\spf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}|;
    }
    else {
      $sTmpProc = qr|$sProc.sap${sSAPSID}_${sInstName}.*pf=.*profile/${sSAPSID}_${sInstName}_${sSAPHost}|;
    }

    @{ $$rhlSrvrDetails { $sProc } } = ( $sPidFile, $sTmpProc ); # Store..

    &logTrace ( 20198, "<%s> Generating filter [%s] for process [%s] with PID file [%s]",
    $sMyName, $sTmpProc, $sProc, $sPidFile );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20199, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  #-----------------------------------------
  # Form query string based on entry point..
  #-----------------------------------------
  $sCommand = $sCmdHares; # Get global..
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
  $sCommandArgs = "-value $sResourceName $sAttribute";

  #----------------------------------
  # Fire command and collect output..
  #----------------------------------
  ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                    =~ s/\n//;

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20200,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
  }
  elsif ( ! $iEngineTimeout ) { # Undefined value..
    &logTrace ( 20201, "<%s> Could not get [%s] from engine",
    $sMyName, $sAttribute );
  }
  else { # Success..
    &logTrace ( 20202, "<%s> Got [%s] as [%s] sec from engine",
    $sMyName, $sAttribute, $iEngineTimeout );
  }

  #-----------------------------------------
  # Form query string based on entry point..
  #-----------------------------------------
  unless ( $iEngineTimeout ) {
    $sCommand = $sCmdHatype; # Get global..
    ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
    $sResourceType = "SAPWebAS71" if $sResourceType !~ /\w/; # Ensure..
    $sCommandArgs = "-value $sResourceType $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20203,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
      $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
    }
    elsif ( ! $iEngineTimeout ) { # Empty value..
      &logTrace ( 20204, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
      $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
    }
    else { # Success..
      &logTrace ( 20205, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20206,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

=pod

 -------------------------------------------------------------------------
 Function : bool UpdateSAPWebAS71Log ( IN String InstName,
                                       IN String InstProfile,
                                       IN String InstType,
                                       IN String SAPAdmin,
                                       IN String SAPSID,
                                       IN String MessageText );

 Description : Update SAP log with passed message string..

 Parameters :

 $sInstName         String identifier that uniquely identifies a SAP server instance
 $sInstProfile      Specifies the full path to the Instance profile for this instance
 $sInstType         String identifier that classifies the SAP server instance type
 $sSAPAdmin         UNIX user dedicated for this instance of SAP
 $sSAPSID           SAP system name defined during installation
 $sMessageText      Message text for update


 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub UpdateSAPWebAS71Log {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstName,
       $sInstProfile,
       $sInstType,
       $sSAPAdmin,
       $sSAPSID,
       $sMessageText ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20207, "Subroutine <%s> called with:
  InstName                 [%s]
  InstProfile              [%s]
  InstType                 [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  MessageText              [%s]",

  $sMyName, $sInstName, $sInstProfile,  $sInstType,
  $sSAPAdmin, $sSAPSID, $sMessageText );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sDate = strftime '%H:%M:%S %d %b %Y', localtime;
  my $sHost = ( uname ) [1];
  my $sLogFile = "/usr/sap/$sSAPSID/$sInstName/log/" .
                 "${sSAPSID}_${sInstName}.log";

  my $sCpyHeader = '+' . '-' x 75 . '+';
  my $sCmnt = "|";

  #------------------------
  # Define logging format..
  #------------------------

  format SAPLOG =
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$sCpyHeader
@ Date: @<<<<<<<<<<<<<<<<<<<<< Node: @<<<<<<<<<<<<<<<<<<                    @
$sCmnt, $sDate, $sHost, $sCmnt
@ InstName: @<<<<<<<<<<<< InstType: @<<<<<<<<<< SAPSID: @<<<                @
$sCmnt, $sInstName, $sInstType, $sSAPSID, $sCmnt
@ Inst profile: @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< @
$sCmnt, $sInstProfile, $sCmnt
@ @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< @
$sCmnt, $sMessageText, $sCmnt
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$sCpyHeader
.

  #-----------------------
  # Open file for update..
  #-----------------------
  if ( sysopen ( SAPLOG, $sLogFile, O_CREAT|O_WRONLY|O_APPEND|O_SYNC, 0644 ) ) {
    &logTrace ( 20208,
    "<%s> Opened log file [%s] successfully for update", $sMyName, $sLogFile );

    write SAPLOG; # Update..

    unless ( close ( SAPLOG ) ) {
      &logError ( 20209,
      "<%s> Failed to close log file [%s] after update: [%s]",
      $sMyName, $sLogFile, $! );
      $iReturnCode = 0; # Failure..
    }
  }
  else {
    &logError ( 20210,
    "<%s> Failed to open log file [%s] for update: [%s]",
    $sMyName, $sLogFile, $! );
    $iReturnCode = 0; # Failure..
  }

  #----------------
  # Set ownership..
  #----------------
  my ( $iUid, $iGid ) = ( getpwnam $sSAPAdmin ) [2, 3]; # Should not fail..

  if ( chown ( $iUid, $iGid, $sLogFile ) ) {
    &logTrace ( 20211,
    "<%s> Set ownership of [%s] to UID [%s] GID [%s]",
    $sMyName, $sLogFile, $iUid, $iGid );
  }
  else {
    &logError ( 20212,
    "<%s> Failed to set ownership of [%s] to UID [%s] GID [%s]: [%s]",
    $sMyName, $sLogFile, $iUid, $iGid, $! );
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExecJSMon ( IN String SAPAdmin,
                            IN String EnvFile,
                            IN String Cmd,
                            IN String CmdArgs,
                            IN Integer Timeout );

 Description : Decides on status of SAP Java Instance based on jsmon command.

 Parameters :

 $sSAPAdmin           UNIX user dedicated for this instance of SAP
 $sEnvFile            Full path to the environment file for SAP
 $sCmd                Full path to jsmon binary
 $sCmdArgs            Argument list for jsmon
 $iTimeout            Time in seconds for execution

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut


sub ExecJSMon {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20213, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iExitCode, @lsCmdOut ) = ();

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20214,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20215,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20216,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  $iReturnCode = SAPWebAS71::ExecLocalCmd ( $sSAPAdmin, # User context..
                                            $sEnvFile,  # Use envfile..
                                            $sCmd,
                                            $sCmdArgs,
                                            \$iExitCode,
                                            $iTimeout,
                                            \@lsCmdOut ); 

  if( defined $iExitCode ){
    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20217,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );

    #------------------------------------------------------
    # Get status based on state of dispatcher and servers..
    #------------------------------------------------------
      $iReturnCode = SAPWebAS71::GetJSMonStatus ( \@lsCmdOut );
      @lsCmdOut = (); # Free..
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20218,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  elsif( $iExitCode == 4 ) {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    &logError ( 20328,
    "<%s> Cmd[%s] executed successfully with exit code 4 indicating  all processes stopped",
    $sMyName, $sCmd );
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logError ( 20219,
    "<%s> Cmd [%s] failed with exit code [%s]. Cmd output [%s]
    Returning [%s]", $sMyName, $sCmd, $iExitCode,
    join ( chr 012, @lsCmdOut ), $iReturnCode );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExecENSMon ( IN String SAPAdmin,
                             IN String EnvFile,
                             IN String Cmd,
                             IN String CmdArgs,
                             IN String InstType,
                             IN Integer Timeout );

 Description : Decides on status of SAP Instance based on ensmon command.

 Parameters :

 $sSAPAdmin   UNIX user dedicated for this instance of SAP
 $sEnvFile    Full path to the environment file for SAP
 $sCmd        Full path to sapinfo binary
 $sCmdArgs    Argument list for ensmon
 $sInstType   String identifier that classifies the SAP server instance type
 $iTimeout    Time in seconds for execution

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExecENSMon {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $sInstType,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20220, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  InstType      [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $sInstType,
  $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sEntryPointName, $iExitCode, @lsCmdOut ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20221,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20222,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20223,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( SAPWebAS71::ExecLocalCmd ( $sSAPAdmin, # User context..
                                     $sEnvFile,  # Use envfile..
                                     $sCmd,
                                     $sCmdArgs,
                                     \$iExitCode,
                                     $iTimeout,
                                     \@lsCmdOut ) ) {

    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20224,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20225,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  elsif ( 4 == $iExitCode && $sInstType =~ /^enqueue$/io ) {
    if ( $sEntryPointName eq 'monitor' ) {
      &logCritical ( 20226,
      "<%s> Cmd [%s] with args [%s] exited with [%s].
      Not able to connect to ENQUEUE Replication Server!!",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
    else {
      &logTrace ( 20227,
      "<%s> Cmd [%s] with args [%s] exited with [%s].
      Not able to connect to ENQUEUE Replication Server!!",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
  }
  elsif ( 8 == $iExitCode ) {  # Up already..
    if ( $sInstType =~ /^enqueue$/io ) {
      $iReturnCode = VCSDef_MonitorExit::OFFLINE;

      &logError ( 20228,
      "<%s> Cmd [%s] with args [%s] exited with [%s]. Returning [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode, $iReturnCode );
    }
    elsif ( $sInstType =~ /^enqrep$/io ) {
      &logCritical ( 20229,
      "<%s> Cmd [%s] with args [%s] exited with [%s].
      Not able to connect to ENQUEUE Server!!",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
  }
  else { # Error..
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logError ( 20230,
    "<%s> Cmd [%s] failed with exit code [%s]. Returning [%s]",
    $sMyName, $sCmd, $iExitCode, $iReturnCode );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExecSAPInfo ( IN String SAPAdmin,
                              IN String EnvFile,
                              IN String Cmd,
                              IN String CmdArgs,
                              IN Integer Timeout );

 Description : Execute sapinfo command in given timeout

 Parameters :

 $sSAPAdmin           UNIX user dedicated for this instance of SAP
 $sEnvFile            Full path to the environment file for SAP
 $sCmd                Full path to sapinfo binary
 $sCmdArgs            Argument list for sapinfo
 $iTimeout            Time in seconds for execution

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut


sub ExecSAPInfo {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20231, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  Timeout       [%s]",

  $sMyName, $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iExitCode, @lsCmdOut ) = ();

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20232,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20233,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20234,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( SAPWebAS71::ExecLocalCmd ( $sSAPAdmin, # User context..
                                     $sEnvFile,  # Use envfile..
                                     $sCmd,
                                     $sCmdArgs,
                                     \$iExitCode,
                                     $iTimeout,
                                     \@lsCmdOut ) ) {

    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20235,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20236,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  elsif ( 2 == $iExitCode ) {  # Server is down..
    $iReturnCode = VCSDef_MonitorExit::OFFLINE;

    &logError ( 20237,
    "<%s> Cmd [%s] with args [%s] exited with [%s]. Cmd output [%s].
    Returning [%s]", $sMyName, $sCmd, $sCmdArgs, $iExitCode,
    join ( chr 012, @lsCmdOut ), $iReturnCode );
  }
  else { # Error..
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logError ( 20238,
    "<%s> Cmd [%s] failed with exit code [%s]. Cmd output [%s]
    Returning [%s]", $sMyName, $sCmd, $iExitCode,
    join ( chr 012, @lsCmdOut ), $iReturnCode );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetJSMonStatus ( IN RefStrList CmdOut );

 Description : Decides on status of SAP Instance based on given jsmon
               output.

 Parameters :

 $rlsCmdOut      Reference to list containing output from jsmon execution.

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetJSMonStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rlsCmdOut ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20239, "Subroutine <%s> called with:
  rlsCmdOut         [%s]",
  $sMyName, $rlsCmdOut );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------------
  # Validate against empty list..
  #------------------------------
  return VCSDef_MonitorExit::UNKNOWN unless scalar @$rlsCmdOut;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sTmp, $sKey ) = ( '' ) x 2;
  my ( $iSrv, $iRunning ) = ( 0 ) x 2;
  my %hsCmdOut = map { /\b(smdagent|server\d*)\b\,\s*\d*(.*)/io } @$rlsCmdOut;

  foreach $sKey ( sort keys %hsCmdOut ) {
    $sTmp .= "Server Name [$sKey] State [$hsCmdOut{$sKey}]\n";
    if ( $sKey =~ /(smdagent|server)/io ) {
      ++ $iSrv;
      ++ $iRunning if $hsCmdOut { $sKey } =~ /Running/io;
    }
  }

  &logTrace ( 20240, "<%s> SAP sapcontrol status : ---\n%s", $sMyName, $sTmp );

  #-----------------
  # Check on state..
  #-----------------
  if ( $iRunning ) {                            # One server running..
    &logTrace ( 20241,
    "<%s> Found [%s] of [%s] server(s) running. Returning [%s]",
    $sMyName, $iRunning, $iSrv, $iReturnCode );
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logWarn ( 20243,
    "<%s> Status of SAP instance undecided:\n%sReturning [%s]",
    $sMyName, $sTmp, $iReturnCode );
  }

  ( $sTmp, $sKey ) = (); # Free..

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool ChkResName ( IN String ResName,
                              IN String SAPSID,
                              IN String InstType );

 Description : Checks if resource has been configured

 Parameters :

 $sResName         Resource Name to check
 $sSAPSID          SAP system name defined during installation
 $sInstType        String identifier that classifies the SAP server instance type

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ChkResName {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sResName,
       $sSAPSID,
       $sInstType ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20244, "Subroutine <%s> called with:
  ResName            [%s]
  SAPSID             [%s]
  InstType           [%s]",

  $sMyName, $sResName, $sSAPSID, $sInstType );
  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1;

  #-------------------------------
  # Validate string for resource..
  #-------------------------------
  return 0 unless ( $sResName =~ /\S/o ); # Must be a non-white string..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHares; # Get global..
  my $sCmdArgs = "-display $sResName -attribute InstProfile InstType";

  #-----------------------
  # Define valid matches..
  #-----------------------
  my %hsInstMatch = ( 'ENQREP'  => 'ENQUEUE' );

  my ( $iExitCode, @lsCmd ) = ();

  &logTrace ( 20245,
  "<%s> Executing Cmd [%s] with Args [%s]",
   $sMyName, $sCmd, $sCmdArgs );
  #--------------------------
  # Fire command and verify..
  #--------------------------
  ( @lsCmd ) = grep {
                 ( s/\n// || 1 )   &&            # Chomp, and make true..
                 /InstProfile|InstType/ &&       # Contains these..
                 s/.*\s+//                       # Get last field..
               } readpipe ( "$sCmd $sCmdArgs" ); # From output..


  my ($localSAPSID, $localInstName, $localSAPHost ) = ();
  my (@lsWords) = ();

  (@lsWords) = split( /\//, $lsCmd [0] );
  ($localSAPSID, $localInstName, $localSAPHost) = split(/_/, $lsWords[-1]);

  #--------------------
  # Collect exit code..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20246,
    "<%s> Cmd [%s] with Args [%s] failed with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ( 2 == scalar @lsCmd )                  &&          # Got both values..
          ( $sSAPSID eq $localSAPSID )            &&          # SAPSID is a match..
          defined $hsInstMatch { uc $sInstType }  &&          # InsType is defined..
          $hsInstMatch { uc $sInstType } eq uc $lsCmd [-1] ) { # And a correct pair..
    &logTrace ( 20247,
    "<%s> Resource [%s] defined in cluster with InstType [%s] SAPSID [%s]",
     $sMyName, $sResName, $lsCmd [-1], $localSAPSID );
  }
  else {
    &logError ( 20248,
    "<%s> Resource [%s] not a matching ENQUEUE variant for this resource.
    The resource is with InstType [%s] SAPSID [%s]",
    $sMyName, $sResName, $lsCmd [-1], $localSAPSID );
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool OnlineENQGrp ( IN String Sys,
                                IN String Grp,
                                IN String Reason,
                                IN String FaultedNode );

 Description : Pre-online trigger for SAPWebAS71

 Parameters :

 $sSys                System where the group is asked to be made online
 $sGrp                Group for which preonline has been invoked
 $sReason             Contains the string MANUAL or FAULT
 $sFaultedNode        Cluster node where the fault occurred due to FAULT

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub OnlineENQGrp {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSys,
       $sGrp,
       $sReason,
       $sFaultedNode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20249, "Subroutine <%s> called with:
  Sys             [%s]
  Grp             [%s]
  Reason          [%s]
  FaultedNode     [%s]",

  $sMyName,
  $sSys, $sGrp, $sReason, $sFaultedNode );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHares; # Get global..

  my %hsInstMatch = ( 'ENQREP'  => 'ENQUEUE' );

  my ( $i, $iExitCode, $sRes, $sDisp, $sGroup, $sNodeName,  $sEnqSrvRes,
       $sCmdArgs,   $sTempNode,    $sEnqRes,   $sInstType,   $sFailHost,
       $sKey, $sSAPSID,  $sSID, @lsHaresOut, @lsCpy, @lsValidNodes, @lsEnqReps,
       @lsEnques,        %hsEnqueues,     %hsFltd,           %hsEnqReps,
       %hsEnqRepGrps, %hsEnqRepHosts, %hsSwEnqReps,    %hsOnlineEnqueues ) = ();

  my (@lsWords, $sInstProfile, $sInstName, $sSAPHost ) = ();
  my $sOSVer = ();

  #---------------------
  # Validate arguments..
  #---------------------
  if ( $sReason eq 'MANUAL' ) {
    &logInfo ( 20250,
    "<%s> Invoked during a normal switch-over or online.", $sMyName );
  }
  elsif ( $sSys =~ /\S/ && ! inet_aton ( $sSys ) ) {
    &logError ( 20251,
    "<%s> Destination cluster node [%s] cannot be resolved",
    $sMyName, $sSys );
    $iReturnCode = 0; # Error..
  }
  elsif ( $sFaultedNode ne 'UNDEF' && $sFaultedNode =~ /\S/ && ! inet_aton ( $sFaultedNode ) ) {
    &logError ( 20252,
    "<%s> Faulted cluster node [%s] cannot be resolved",
    $sMyName, $sFaultedNode );
    $iReturnCode = 0; # Error..
  }
  elsif ( ! -e $sCmd ) {
    &logError ( 20253,
    "<%s> Command [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Error..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20254,
    "<%s> Command [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Error..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------------------------------------------
  # Bug in VCS for FAULT. Verify group is really faulted..
  #-------------------------------------------------------
  # hasys -list SysState=RUNNING
  $sCmd = $sCmdHasys;
  $sCmdArgs = "-state";

  my ( @lsHasysRunning, @lsHasysFault, @lsHasysOut, %hsHasysRunning, %hsHasysFault ) = ();

  @lsHasysOut = readpipe ( "$sCmd $sCmdArgs" );
  shift @lsHasysOut;

  @lsHasysRunning = grep {
                  ( s/\n//  || 1 ) &&
                  ( s/.*:// || 1 ) &&   # for DR ...
                  ( s/^(\S+).*(RUNNING)$/$1/ )
                } @lsHasysOut;

  @lsHasysFault = grep {
                  ( s/\n//  || 1 ) &&
                  ( s/.*:// || 1 ) &&   # for DR ...
                  ( s/^(\S+).*(FAULTED|EXITED)$/$1/ )
                } @lsHasysOut;

  %hsHasysRunning = map { $_ => 1 } @lsHasysRunning;

  &logTrace ( 20255,
  "<%s> List of systems where VCS running [%s]", 
  $sMyName, join ( ' ', @lsHasysRunning ) );

  %hsHasysFault   = map { $_ => 1 } @lsHasysFault;

  &logTrace ( 20256,
  "<%s> List of systems where VCS Faulted [%s]", 
  $sMyName, join ( ' ', @lsHasysFault ) );

  $sCmd = $sCmdHares;
  $sCmdArgs = "-list Group=$sGrp State=FAULTED";

  @lsHaresOut = grep {
                  ( s/\n// || 1 ) &&
                  s/\s+.*//
                } readpipe ( "$sCmd $sCmdArgs" );

  #-------------------
  # Verify exit code..
  #-------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20257,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
    Group [%s] does not have any faulted resources.",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, $sGrp );

    #-----------------------------
    # Check for node FAULT event..
    #-----------------------------
    if ( $sFaultedNode eq 'UNDEF' && $sReason eq 'FAULT' ) {
      &logWarn ( 20258,
      "<%s> Missing faulted node name.
      Possible node FAULT event detected", $sMyName );
    }
    elsif ( $sFaultedNode eq 'UNDEF' && $sReason eq 'MANUAL' ) {
      &logWarn ( 20259,
      "<%s> Missing source node name.", $sMyName );
    }
    elsif ( ($sOSVer = $^O eq 'solaris' ? ( uname ) [2] : '') eq "5.10" ) {
      #---------------------------------------------------------------
      # Add zone specific support for Enqueue when zone itself faults.
      #---------------------------------------------------------------
      &logInfo ( 20260,
      "<%s> Checking if the Group [%s] is configured for Solaris non-global zones",
      $sMyName, $sGrp );

      my $sOut = ();
      my @lsOut = ();
      my $iCount = ();

      $sCmdArgs = "-display -attribute ContainerName -group $sGrp -type SAPWebAS71";
      @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCmd $sCmdArgs" );

      shift @lsHaresOut;

      foreach $sOut (@lsHaresOut) {
        @lsOut = split( '\s+', $sOut);
        $iCount = scalar @lsOut;
        if( (4 == $iCount) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
          &logWarn ( 20261,
          "<%s> ContainerName attribute is configured.
          Possible zone [%s] FAULT event detected", $sMyName, $lsOut[3] );

          last;
        }
      }
    }
  }
  else {
    &logTrace ( 20262,
    "<%s> Group [%s] has [%s] resource(s) in FAULTED state",
    $sMyName, $sGrp, join ( ' ', @lsHaresOut ) );
  }

  #----------------
  # Re-initialise..
  #----------------
  $sCmdArgs = "-display -type SAPWebAS71 -attribute " .
              "State Group EnqSrvResName InstProfile InstType";

  #---------------------------------------------
  # Assume the group name is valid and proceed..
  #---------------------------------------------
  @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sCmd $sCmdArgs" );

  #-----------------------------------------
  # Check if command executed successfully..
  #-----------------------------------------
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20263,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    return 0; # Error..
  }
  elsif ( $lsHaresOut[0] =~ /10554.*No resource exists with/o ) {
    &logError ( 20264,
    "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
    $sMyName, $sCmd, $sCmdArgs,  $lsHaresOut[0] );
    return 0; # Error..
  }

  $sFailHost = $sSys;    # Set to default..
  push @lsHaresOut, "#"; # Insert trailing de-marker..
  @lsCpy = @lsHaresOut;  # Backup..

  #---------------------------------------------------
  # Mark all ENQUEUE/ENQREP resources in given group..
  #---------------------------------------------------
  @lsEnques =
  grep {
     ( s/\n// || 1 )                   && # Chomp, and always make true..
     ( /Group.*\s+(\S+)/               && # Mark the group name..
     ( $sGroup = $1 ) && 0 )           || # Store, and always make false..
     ( m#InstProfile.*\s+.*/(...)#     && # Mark the SID..
     ( $sSID = $1 ) && 0 )             || # Store, and always make false..
     ( s/^(\S+).*InstType.*\s             # Mark resource..
     (ENQ(UEUE|REP))/$1:$2/ix          && # If ENQUEUE or ENQREP variant..
     ++ $i )                           || # Set flag..
     ( $i                              && # If flag..
     s/.*/sprintf ( "$sSID:$sGroup",      # Print group..
      -- $i )/e )                         # Reset flag..
  } @lsCpy;

  &logTrace ( 20265,
  "<%s> Searching ENQUEUE resource in group [%s]", $sMyName, $sGrp );

  &logTrace ( 20266,
  "<%s> Enques values are [%s]", $sMyName, join(' ', @lsEnques) );

  #-------------------
  # Start processing..
  #-------------------
  while ( 2 == ( ( $sRes, $sGroup ) = splice ( @lsEnques, 0, 2 ) ) ) {
    ( $sRes, $sInstType ) = $sRes =~ /([^:]+)/g;
    ( $sSAPSID, $sGroup ) = $sGroup =~ /([^:]+)/g;

    #-------------------------------
    # Passed group has ENQ(UE|REP)..
    #-------------------------------
    if ( $sGrp eq $sGroup ) {
      if ( $sInstType =~ /ENQUEUE/io ) {
        &logInfo ( 20267,
        "<%s> +------- Group [%s] has ENQUEUE variant -------+",
        $sMyName, $sGrp );

        #--------
        # Store..
        #--------
        push @{ $hsEnqueues { $sRes } },
        $sSAPSID, $sInstType, $sGroup; # Store..
      }
      elsif ( $sInstType =~ /ENQREP/io ) {
        &logInfo ( 20268,
        "<%s> +------- Group [%s] has ENQREP variant -------+",
        $sMyName, $sGrp );

        #--------
        # Store..
        #--------
        push @{ $hsSwEnqReps { $sRes } },
        $sSAPSID, $sInstType, $sGroup; # Store..
      }
    }

    &logTrace ( 20269,
    "<%s> Resource [%s] SAPSID [%s] InstType [%s] found in Group [%s]",
    $sMyName, $sRes, $sSAPSID, $sInstType, $sGroup );
  }

  #-----------------------------------------------
  # Find online resources and list their details..
  #-----------------------------------------------
  @lsCpy = @lsHaresOut; # Re-initialise..
  $i = 0;

  &logTrace ( 20270,
  "<%s> Looking for all ONLINE SAPWebAS71 resources", $sMyName );

  @lsEnqReps =
  grep {
    ( s/\n// || 1 )                        && # Chomp, and always make true..
    ( /^(\S+).*Group.*\s+(\S+)/            && # Mark resource, and group name..
    ( $sGroup = "$1:$2" ) && 0 )           || # Store, and always make false..
    ( s/.*\s(\S+)\s+ONLINE$/$1/ && ++ $i ) || # If online, get node name, set flag..
    ( $i                                   && # If flag..
    /EnqSrvResName|InstProfile|InstType/   && # and contains one of these..
    s/.*\s+// )                            || # get the last field..
    ( $i                                   && # If flag, use demarker..
    s/#/sprintf ( $sGroup, -- $i )/e )        # to print the group name.. reset flag..
  } @lsCpy;                                   # From the copy..

  &logTrace ( 20271,
  "<%s> EnqReps [%s]", $sMyName, join(' ', @lsEnqReps) );

  while ( 5 == ( ( $sNodeName, $sEnqSrvRes, $sInstProfile, $sInstType, $sGroup ) = splice ( @lsEnqReps, 0, 5 ) ) ) {

    $sNodeName =~ s/.*://; # Remove cluster component, under GCO..
    ( $sRes, $sGroup ) = $sGroup =~ /([^:]+)/g;

    #----------------------------------
    # Store only ENQ(UE|REP) variants..
    #----------------------------------
    (@lsWords) = split( /\//, $sInstProfile);
    ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);

    #----------------------------------
    # Store only ENQ(UE|REP) variants..
    #----------------------------------
    if ( $sInstType =~ /ENQREP/io ) {
      &logTrace ( 20272, "<%s> Located ENQREP variant. Resource [%s] details:
      Group [%s] Node [%s] EnqSrvResName [%s] InstType [%s] SAPSID [%s]",
      $sMyName, $sRes, $sGroup, $sNodeName, $sEnqSrvRes, $sInstType, $sSAPSID );

      push @{ $hsEnqReps { $sRes } },
      $sGroup, $sEnqSrvRes, $sInstType, $sSAPSID, $sNodeName;
    }
    elsif ( $sInstType =~ /ENQUEUE/io ) {
      $sEnqSrvRes = ""; # Truncate..
      &logTrace ( 20273, "<%s> Located ENQUEUE variant. Resource [%s] details:
      Group [%s] Node [%s] EnqSrvRes [%s] InstType [%s] SAPSID [%s]",
      $sMyName, $sRes, $sGroup, $sNodeName, $sEnqSrvRes, $sInstType, $sSAPSID );

      push @{ $hsOnlineEnqueues { $sRes } },
      $sGroup, $sEnqSrvRes, $sInstType, $sSAPSID, $sNodeName;
    }
  }

  #--------------------------------------------
  # Validate if group has ENQ(UE|REP) variant..
  #--------------------------------------------
  @lsCpy = @lsHaresOut; # Re-initialise.

  if ( keys %hsEnqueues ) {
    &logInfo ( 20274,
    "<%s> Failover Group [%s] has ENQUEUE variant resource(s) [%s].",
    $sMyName, $sGrp, join ( ' ' , keys %hsEnqueues ) );
  }
  elsif ( keys %hsSwEnqReps ) {
    #------------------------------------------
    # Get all valid nodes for the faulted REP..
    #------------------------------------------
    @lsCpy = @lsHaresOut;          # Re-initialise.
    $sRes =  join ( '|', keys %hsSwEnqReps );
    $sDisp =  join ( ' ', keys %hsSwEnqReps );

    &logInfo ( 20275,
    "<%s> Failover Group [%s] has ENQREP variant resource(s) [%s]",
    $sMyName, $sGrp, $sDisp );

    %hsFltd = map { $_ => 1 }                 # All faulted nodes..
    grep {                                    # Grep'ed here..
      ( s/\n// || 1 )                    &&   # Chomp, and always make true..
      s/^($sRes).*State\s+(\S+).*FAULTED/$2/  # Pick all nodes where FAULTed..
    } @lsCpy;                                 # From cmd output..

    @lsCpy = @lsHaresOut;                     # Re-initialise.
    @lsValidNodes =
    grep {
      ( s/\n// || 1 )                       &&  # Chomp, and always make true..
      s/^($sRes).*State\s+(\S+).*LINE/$2/   ||  # Pick the node not FAULTed..
      s/^($sRes).*EnqSrvResName.*\s+(\S+)/::$2/ # And EnqSrvResName..
    } @lsCpy;                                   # From cmd output..

    #-------------------------------------------------
    # Now extract EnqSrvResName and configured nodes..
    #-------------------------------------------------
    @lsCpy = @lsValidNodes; # Re-use..
    @lsValidNodes = ();     # Truncate..

    foreach $sKey ( @lsCpy ) {
      if ( $sKey =~ /^::/ ) { # EnqResName found..
        ( $sEnqRes = $sKey ) =~ s/:://;
      }
      else {
        $sKey =~ s/.*://; # Remove cluster component, under GCO..
        push ( @lsValidNodes, $sKey ) unless $hsFltd { $sKey };
      }
    }

    &logTrace ( 20276,
    "<%s> Nodes on which Service Group not in FALUTED state [%s]",
    $sMyName, join(' ', @lsValidNodes) );

    ( $sRes ) = keys %hsSwEnqReps; # Take one..

    #--------------------
    # Remove duplicates..
    #--------------------
    %hsFltd = (); # Truncate..
    @lsValidNodes = grep { ! $hsFltd { $_ } ++ } @lsValidNodes;

    &logTrace ( 20277,
    "<%s> One of ENQREP resource(s) [%s] defines EnqSrvResName as [%s].

    Details for ENQREP resource [%s]:

    InstType [%s] SAPSID [%s]. Possible failover nodes [%s]",
    $sMyName, $sDisp, $sEnqRes, $sRes,
    ${ $hsSwEnqReps { $sRes } } [1], ${ $hsSwEnqReps { $sRes } } [0],
    join ( ' ', @lsValidNodes ) );

    #--------------------------------------------------------
    # Check if the node on which service group is configured 
    # is Running at this moment..
    #--------------------------------------------------------
    @lsValidNodes = grep { $hsHasysRunning { $_ } } @lsValidNodes;

    &logTrace ( 20278,
    "<%s> Valid nodes for ENQREP failover targets [%s]",
    $sMyName, join ( ' ', @lsValidNodes ) );

    #--------------------------
    # If online ENQUEUE found..
    #--------------------------
    if ( defined $hsOnlineEnqueues { $sEnqRes } ) {
      &logTrace ( 20279,
      "<%s> Details for defined ENQUEUE resource [%s]:
      InstType [%s] SAPSID [%s] online on node [%s]",
      $sMyName, $sEnqRes, ${ $hsOnlineEnqueues { $sEnqRes } } [2],
      ${ $hsOnlineEnqueues { $sEnqRes } } [-2],
      ${ $hsOnlineEnqueues { $sEnqRes } } [-1] );
    }
    else {
      &logWarn ( 20280,
      "<%s> Corresponding ENQUEUE resource(s) is NOT online in the cluster", $sMyName );
    }

    #-----------------------------
    # Now validate EnqSrvResName..
    #-----------------------------
    if ( defined $hsOnlineEnqueues { $sEnqRes }             && # Defined..
         ( $hsInstMatch { ${ $hsSwEnqReps { $sRes } } [1] } eq # InstType pair..
         ${ $hsOnlineEnqueues { $sEnqRes } } [2] )          && # is a match..
         ( ${ $hsSwEnqReps { $sRes } } [0]                  eq # SAPSID..
           ${ $hsOnlineEnqueues { $sEnqRes } } [-2] ) ) {      # matches..

      $sTempNode = ${ $hsOnlineEnqueues { $sEnqRes } } [-1];   # Get the online node..

      ( $sFailHost ) = grep {                                  # Find the node other than..
                         ! /^($sTempNode|$sFaultedNode)$/o &&  # ENQUEUE online or faulted..
                         ! $hsHasysFault{$_}                   # nodes faulted..
                       } @lsValidNodes;                        # From all posibilities..

      #---------------------------------------------------
      # Check if we are in a position to failover safely..
      #---------------------------------------------------
      if ( $sFaultedNode ne 'UNDEF' || $sReason ne 'MANUAL' ||
           ( $sFaultedNode eq 'UNDEF' && $sReason eq 'MANUAL' && (scalar @lsHasysFault || $sTempNode eq $sSys)) ) {
        if ( defined $sFailHost ) {
          &logInfo ( 20281,
          "<%s> Rejecting ENQUEUE node [%s]. Failing over group [%s] to node [%s]",
          $sMyName, $sTempNode, $sGrp, $sFailHost );
        }
        else { # Prevent onlining..
          &logError ( 20282,
          "<%s> Could not locate any node to online Group [%s]",
          $sMyName, $sGrp );
          return $iReturnCode; # Error.. Prevent online of ENQREP..
        }
      }

      #-------------------------------
      # Online group to default node..
      #-------------------------------
      if ( SAPWebAS71::OnlineGrp ( $sGrp,
                                   $sFailHost ) ) {
        &logInfo ( 20283,
        "<%s> Issued online of Group [%s] on node [%s]",
        $sMyName, $sGrp, $sFailHost );
        return $iReturnCode; # Success..
      }
      else {
        &logError ( 20284,
        "<%s> Could not issue online of Group [%s] on node [%s]",
        $sMyName, $sGrp, $sFailHost );
        return 0; # Error..
      }
    }
    else {
      &logTrace ( 20285,
      "<%s> Failover group [%s] does not have any online ENQ(UEUE|REP) variant",
      $sMyName, $sGrp );
      return 0; # Do nothing..
    }
  }
  else {
    &logInfo ( 20286,
    "<%s> Failover group [%s] does not have any ENQ(UEUE|REP) variant",
    $sMyName, $sGrp );
    return 0; # Do nothing..
  }

  #------------------------------------------------
  # Validate if we have any online ENQREP variant..
  #------------------------------------------------
  &logTrace ( 20287,
  "<%s> +----- ENQREP online node and group details -----+", $sMyName );

  #----------------------------------------------
  # Remove ENQREPs whose EnqRepSrv aren't found..
  #----------------------------------------------
  foreach $sKey ( keys %hsEnqReps ) {
    if ( ! $hsEnqueues { ${$hsEnqReps { $sKey } }[1] } ) { # Not in passed group..
      &logTrace ( 20288,
      "<%s> EnqResName [%s] resource not found in Grp [%s].
      Marked resource [%s] invalid",
      $sMyName, ${$hsEnqReps { $sKey } }[1], $sGrp, $sKey );

      delete $hsEnqReps { $sKey };                        # Then delete..
    }
    else {
      ++ $hsEnqRepHosts { ${$hsEnqReps { $sKey } }[-1] }; # Else store online node..
      ++ $hsEnqRepGrps  { ${$hsEnqReps { $sKey } }[0] };  # Else store group..

      &logTrace ( 20289,
      "<%s> Online node [%s] Group [%s]",
      $sMyName, ${$hsEnqReps{$sKey}}[-1], ${$hsEnqReps{$sKey}}[0] );
    }
  }

  #------------------------------------------------
  # Validate if we have any online ENQREP variant..
  #------------------------------------------------
  if ( ! keys %hsEnqReps && $sReason ne 'MANUAL' ) {
    $sFailHost = $sSys; # Fail-over to the default node..
    &logCritical ( 20290,
    "<%s> Corresponding ENQREP server(s) not found ONLINE in cluster.
    Onlining group [%s] on node [%s]", $sMyName, $sGrp, $sFailHost );

    #-------------------------------
    # Online group to default node..
    #-------------------------------
    if ( SAPWebAS71::OnlineGrp ( $sGrp,
                                 $sFailHost ) ) {
      &logInfo ( 20291,
      "<%s> Issued online of Group [%s] on node [%s]",
      $sMyName, $sGrp, $sFailHost );
      return $iReturnCode; # Success..
    }
    else {
      &logError ( 20292,
      "<%s> Could not issue online of Group [%s] on node [%s]",
      $sMyName, $sGrp, $sFailHost );
      return 0; # Error..
    }
  }

  #-----------------------------------------------
  # Check if we have ENQREPs in different groups..
  #-----------------------------------------------
  if ( keys %hsEnqRepGrps > 1 ) {
    &logCritical ( 20293,
    "<%s> ENQREP configured in different service groups [%s]",
    $sMyName, join ( ' ', keys %hsEnqRepGrps ) );
  }

  #----------------------------------------------
  # Now check if we have a unique failover node..
  #----------------------------------------------
  if ( keys %hsEnqRepHosts > 1 ) {
    &logCritical ( 20294,
    "<%s> ENQREP found online on different nodes [%s]",
    $sMyName, join ( ' ', keys %hsEnqRepHosts ) );

    #------------------------------------------
    # Delete entries for all other than ENQREP..
    #------------------------------------------
    foreach $sKey ( keys %hsEnqReps ) {
      if ( ${$hsEnqReps { $sKey } }[2] ne 'ENQREP' ) {
        &logError ( 20295,
        "<%s> Marking resource [%s] with InstType [%s] as invalid",
        $sMyName, $sKey, ${$hsEnqReps { $sKey } }[2] );

        #--------------------------
        # Delete in correct order..
        #--------------------------
        delete $hsEnqRepHosts { ${$hsEnqReps { $sKey } }[-1] }; # Delete node first..
        delete $hsEnqRepGrps  { ${$hsEnqReps { $sKey } }[0] };  # Delete group next..
        delete $hsEnqReps { $sKey };                            # Now the resource..
      }
    }
  }

  #-------------------------
  # Find out failover node..
  #-------------------------
  foreach $sKey ( keys %hsEnqReps ) {
    &logTrace ( 20296,
    "<%s> EnqRep server [%s] details:
    Group [%s] EnqSrvRes [%s] InstType [%s] SAPSID [%s] online on [%s]",
    $sMyName, $sKey,
    ${ $hsEnqReps { $sKey } }[0], ${ $hsEnqReps { $sKey } }[1],
    ${ $hsEnqReps { $sKey } }[2], ${ $hsEnqReps { $sKey } }[3],
    ${ $hsEnqReps { $sKey } }[4] );

    #-----------------------------------------------------------
    # Enqueue has been defined in EnqResName of ENQREP variant..
    #-----------------------------------------------------------
    if ( scalar @{ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } } ) { # EnqResName is correct..
      if ( ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[0] ne    # SAPSID of ENQUEUE variant..
                            ${$hsEnqReps { $sKey }}[3] ) {          # ne SAPSID of ENQREP..
        &logError ( 20297,
        "<%s> SAPSID [%s] of ENQREP resource [%s] differs from
        SAPSID [%s] of ENQUEUE resource [%s]",
        $sMyName, ${$hsEnqReps { $sKey }}[3], $sKey,
        ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[0],
        ${ $hsEnqReps { $sKey } }[1] );
      }
      elsif ( uc ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[1] ne # ENQUEUE variant..
              $hsInstMatch { uc ${$hsEnqReps {$sKey }}[2] } ) {        # not correct match..
        &logError ( 20298,
        "<%s> InstType [%s] of ENQREP resource [%s] not a match for
        InstType [%s] of ENQUEUE resource [%s]", $sMyName,
        ${$hsEnqReps { $sKey }}[2], $sKey,
        ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[1],
        ${ $hsEnqReps { $sKey } }[1] );
      }
      else { # All okay..
        &logTrace ( 20299,
        "<%s> EnqSrvResName [%s] matches with:
        ENQUEUE SAPSID [%s] InsType [%s] Group [%s]", 
        $sMyName, ${$hsEnqReps { $sKey }}[1], 
        ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[0],
        ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[1],
        ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[2] );

        #-----------------------------------------------------
        # Failover to ENQREP's node, if not the FAULTED node..
        #-----------------------------------------------------
        $sFailHost = ${ $hsEnqReps { $sKey } }[-1] eq $sFaultedNode ? $sSys :
                     ${ $hsEnqReps { $sKey } }[-1];
      }
    }
    else { # EnqResName is wrong..
      &logError ( 20300,
      "<%s> EnqResName [%s] value is in-correct for ENQREP resource [%s]",
      $sMyName,  ${$hsEnqReps { $sKey }}[1], $sKey );
    }
  }

  &logInfo ( 20301, "<%s> Failing over Group [%s] to node [%s]",
  $sMyName, $sGrp, $sFailHost );

  #-------------------------------
  # Online group to default node..
  #-------------------------------
  if ( SAPWebAS71::OnlineGrp ( $sGrp,
                               $sFailHost ) ) {
    &logInfo ( 20302, "<%s> Issued online of Group [%s] on node [%s]",
    $sMyName, $sGrp, $sFailHost );
  }
  else {
    &logError ( 20303,
    "<%s> Could not issue online of Group [%s] on node [%s]",
    $sMyName, $sGrp, $sFailHost );
    $iReturnCode = 0; # Do nothing..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool OnlineGrp ( IN String Grp,
                             IN String Sys );

 Description : Pre-online trigger for SAPWebAS71

 Parameters :

 $sGrp                Cluster group to online
 $sSys                System where the group is asked to be made online

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub OnlineGrp {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sGrp,
       $sSys ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20304, "Subroutine <%s> called with:
  Grp             [%s]
  Sys             [%s]",

  $sMyName, $sGrp, $sSys );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHagrp; # Get global..
  my @lsCmdArgs = ( "-online", "-nopre", "$sGrp", "-sys", "$sSys" );
  my $iExitCode = undef;

  &logTrace ( 20305, "<%s> Executing Command [%s] with arguments [%s]",
  $sMyName, $sCmd, join( ' ', @lsCmdArgs ) );

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ! -e $sCmd ) {
    &logError ( 20306,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20307,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  else { # Execute..
    &logTrace ( 20308,
    "<%s> Executing Cmd [%s] with args [%s]",
    $sMyName, $sCmd, join ( ' ', @lsCmdArgs ) );

    system ( $sCmd, @lsCmdArgs ); # Execute..

    #-------------------------
    # Check execution status..
    #-------------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20309,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
      $sMyName, $sCmd, join ( ' ', @lsCmdArgs ), $iExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20310,
      "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCmd, join ( ' ', @lsCmdArgs ), $iExitCode );
    }
  }

  return $iReturnCode;
}

1; # So PERL knows its included..
